# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-many-public-methods, too-many-lines

from __future__ import annotations

import json
import re
import uuid
from copy import deepcopy
from typing import TYPE_CHECKING
from warnings import warn

import h5py
import numpy as np

from ..data import (
    CommentsData,
    Data,
    DataType,
    FilenameData,
    GeometricDataValueMapType,
    ReferenceDataType,
    ReferencedData,
    ReferenceValueMap,
)
from ..groups import Group, GroupType, PropertyGroup, RootGroup
from ..objects import ObjectBase, ObjectType
from ..shared import FLOAT_NDV, Entity, EntityType, fetch_h5_handle
from ..shared.concatenation import Concatenator
from ..shared.utils import KEY_MAP, as_str_if_uuid, dict_mapper
from .utils import str_from_subtype, str_from_type


if TYPE_CHECKING:
    from .. import shared, workspace


class H5Writer:
    """
    Writing class to a geoh5 file.
    """

    str_type = h5py.special_dtype(vlen=str)

    @staticmethod
    def init_geoh5(
        file: str | h5py.File,
        workspace: workspace.Workspace,
    ):
        """
        Add the geoh5 core structure.

        :param file: Name or handle to a geoh5 file.
        :param workspace: :obj:`~geoh5py.workspace.workspace.Workspace` object
            defining the project structure.

        :return h5file: Pointer to a geoh5 file.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            project = h5file.create_group(workspace.name)
            H5Writer.write_attributes(h5file, workspace)
            project.create_group("Data")
            project.create_group("Groups")
            project.create_group("Objects")
            types = project.create_group("Types")
            types.create_group("Data types")
            types.create_group("Group types")
            types.create_group("Object types")

    @staticmethod
    def create_dataset(entity_handle, dataset: np.ndarray, label: str) -> None:
        """
        Create a dataset on geoh5.

        :param entity_handle: Pointer to a hdf5 group
        :param dataset: Array of values to be written
        :param label: Name of the dataset on file
        """
        entity_handle.create_dataset(
            label,
            data=dataset,
            dtype=dataset.dtype,
            compression="gzip",
            compression_opts=9,
        )

    @staticmethod
    def remove_child(
        file: str | h5py.File,
        uid: uuid.UUID,
        ref_type: str,
        parent: Entity,
    ) -> None:
        """
        Remove a child from a parent.

        :param file: Name or handle to a geoh5 file
        :param uid: uuid of the target :obj:`~geoh5py.shared.entity.Entity`
        :param ref_type: Input type from: 'Types', 'Groups', 'Objects' or 'Data
        :param parent: Remove entity from parent.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            uid_str = as_str_if_uuid(uid)
            parent_handle = H5Writer.fetch_handle(h5file, parent)

            if parent_handle is None or parent_handle.get(ref_type) is None:
                return

            if uid_str in parent_handle[ref_type]:
                del parent_handle[ref_type][uid_str]
                parent.workspace.repack = True

    @staticmethod
    def remove_entity(
        file: str | h5py.File,
        uid: uuid.UUID,
        ref_type: str,
        parent: Entity | None = None,
    ) -> None:
        """
        Remove an entity and its type from the target geoh5 file.

        :param file: Name or handle to a geoh5 file
        :param uid: uuid of the target :obj:`~geoh5py.shared.entity.Entity`
        :param ref_type: Input type from: 'Types', 'Groups', 'Objects' or 'Data
        :param parent: Remove entity from parent.

        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            base = list(h5file)[0]
            base_type_handle = h5file[base][ref_type]
            uid_str = as_str_if_uuid(uid)

            if ref_type == "Types":
                for e_type in ["Data types", "Group types", "Object types"]:
                    if uid_str in base_type_handle[e_type]:
                        del base_type_handle[e_type][uid_str]
            else:
                if uid_str in base_type_handle:
                    del base_type_handle[uid_str]

                if parent is not None:
                    H5Writer.remove_child(h5file, uid, ref_type, parent)

    @staticmethod
    def fetch_handle(
        file: str | h5py.File,
        entity,
        return_parent: bool = False,
    ) -> None | h5py.Group:
        """
        Get a pointer to an :obj:`~geoh5py.shared.entity.Entity` in geoh5.

        :param file: Name or handle to a geoh5 file
        :param entity: Target :obj:`~geoh5py.shared.entity.Entity`
        :param return_parent: Option to return the handle to the parent entity.

        :return entity_handle: HDF5 pointer to an existing entity, parent or None if not found.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            base = list(h5file)[0]
            base_handle = h5file[base]

            if entity.name == base:
                return base_handle

            uid = entity.uid
            hierarchy = {
                Data: "Data",
                ObjectBase: "Objects",
                Group: "Groups",
                DataType: "Data types",
                ObjectType: "Object types",
                GroupType: "Group types",
            }

            if isinstance(entity, EntityType):
                if "Types" in base_handle:
                    base_handle = base_handle["Types"]
                else:
                    base_handle = base_handle.create_group("Types")

            for key, value in hierarchy.items():
                if isinstance(entity, key):
                    if value in base_handle:
                        base_handle = base_handle[value]
                    else:
                        base_handle = base_handle.create_group(value)

            # Check if already in the project
            if as_str_if_uuid(uid) in base_handle:
                if return_parent:
                    return base_handle

                return base_handle[as_str_if_uuid(uid)]

        return None

    @staticmethod
    def save_entity(
        file: str | h5py.File,
        entity,
        compression: int = 5,
        add_children: bool = True,
    ) -> h5py.Group:
        """
        Save a :obj:`~geoh5py.shared.entity.Entity` to geoh5 with its
        :obj:`~geoh5py.shared.entity.Entity.children` recursively.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target :obj:`~geoh5py.shared.entity.Entity`.
        :param compression: Compression level for the data.
        :param add_children: Add :obj:`~geoh5py.shared.entity.Entity.children`.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            new_entity = H5Writer.write_entity(h5file, entity, compression)

            if (
                add_children
                and not isinstance(entity, Concatenator)
                and hasattr(entity, "children")
            ):
                # Write children entities and add to current parent
                for child in entity.children:
                    if not isinstance(child, PropertyGroup):
                        H5Writer.save_entity(h5file, child, compression)

            H5Writer.write_to_parent(
                h5file, entity, compression=compression, recursively=False
            )

        return new_entity

    @staticmethod
    def update_concatenated_field(
        file: str | h5py.File, entity, attribute: str, channel: str
    ) -> None:
        """
        Update the attributes of a concatenated :obj:`~geoh5py.shared.entity.Entity`.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target :obj:`~geoh5py.shared.entity.Entity`.
        :param attribute: Name of the attribute to get updated.
        :param channel: Name of the data or index to be modified.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle = H5Writer.fetch_handle(h5file, entity)

            if entity_handle is None:
                return

            attr_handle = entity_handle["Concatenated Data"].get(attribute.capitalize())

            if attr_handle is None:
                attr_handle = entity_handle["Concatenated Data"].create_group(
                    attribute.capitalize()
                )
            name = channel.replace("/", "\u2044")
            if name in attr_handle:
                del attr_handle[name]
                entity.workspace.repack = True

            values = getattr(entity, attribute).get(channel, None)

            if values is None:
                return

            if isinstance(values, np.ndarray):
                if np.issubdtype(values.dtype, np.floating):
                    values = values.astype(np.float32)

                    if len(values) > 0:
                        values[np.isnan(values)] = FLOAT_NDV

                if np.issubdtype(values.dtype, np.str_):
                    values = values.astype(h5py.special_dtype(vlen=str))

            attr_handle.create_dataset(
                name,
                data=values,
                compression="gzip",
                compression_opts=9,
            )

    @staticmethod
    def update_field(
        file: str | h5py.File,
        entity,
        attribute: str,
        compression: int = 5,
        **kwargs,
    ) -> None:
        """
        Update the attributes of an :obj:`~geoh5py.shared.entity.Entity`.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target :obj:`~geoh5py.shared.entity.Entity`.
        :param attribute: Name of the attribute to get updated.
        :param compression: Compression level for the data.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle = H5Writer.fetch_handle(h5file, entity)

            if entity_handle is None:
                return
            if attribute == "concatenated_attributes":
                H5Writer.write_group_values(
                    h5file, entity, attribute, compression, **kwargs
                )
            elif attribute in [
                "values",
            ]:
                H5Writer.write_data_values(
                    h5file, entity, attribute, compression, **kwargs
                )
            elif attribute in ["metadata", "options"]:
                H5Writer.write_metadata(h5file, entity, attribute, **kwargs)
            elif attribute in [
                "cells",
                "concatenated_object_ids",
                "layers",
                "octree_cells",
                "property_group_ids",
                "prisms",
                "surveys",
                "trace",
                "trace_depth",
                "u_cell_delimiters",
                "v_cell_delimiters",
                "vertices",
                "z_cell_delimiters",
            ]:
                H5Writer.write_array_attribute(h5file, entity, attribute, **kwargs)
            elif attribute == "property_groups":
                H5Writer.write_property_groups(h5file, entity)
            elif attribute == "color_map":
                H5Writer.write_color_map(h5file, entity)
            elif attribute == "value_map":
                H5Writer.write_value_map(h5file, entity)
            elif attribute == "data_map":
                H5Writer.write_data_map(h5file, entity)
            elif attribute == "entity_type":
                del entity_handle["Type"]
                entity.workspace.repack = True
                new_type = H5Writer.write_entity_type(h5file, entity.entity_type)
                entity_handle["Type"] = new_type
            else:
                H5Writer.write_attributes(h5file, entity)

    @staticmethod
    def write_attributes(
        file: str | h5py.File,
        entity,
    ) -> None:
        """
        Write attributes of an :obj:`~geoh5py.shared.entity.Entity`.

        :param file: Name or handle to a geoh5 file.
        :param entity: Entity with attributes to be added to the geoh5 file.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle = H5Writer.fetch_handle(h5file, entity)

            if entity_handle is None:
                return

            for key, attr in entity.attribute_map.items():
                try:
                    value = getattr(entity, attr)
                except AttributeError:
                    continue

                value = as_str_if_uuid(value)

                if (
                    key
                    in [
                        "PropertyGroups",
                        "Attributes",
                        "Attributes Jsons",
                        "Property Groups IDs",
                        "Concatenated object IDs",
                    ]
                    or value is None
                ):  # or key in Concatenator._attribute_map:
                    continue

                if key in ["Association", "Primitive type"]:
                    value = KEY_MAP[value.name]

                if isinstance(value, (np.int8, bool)):
                    entity_handle.attrs.create(key, int(value), dtype="int8")
                elif isinstance(value, str):
                    entity_handle.attrs.create(key, value, dtype=H5Writer.str_type)
                else:
                    entity_handle.attrs.create(
                        key, value, dtype=np.asarray(value).dtype
                    )

    @staticmethod
    def write_color_map(
        file: str | h5py.File,
        entity_type: shared.EntityType,
    ) -> None:
        """
        Add :obj:`~geoh5py.data.color_map.ColorMap` to a
        :obj:`~geoh5py.data.data_type.DataType`.

        :param file: Name or handle to a geoh5 file
        :param entity_type: Target entity_type with color_map
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            color_map = getattr(entity_type, "color_map", None)
            entity_type_handle = H5Writer.fetch_handle(h5file, entity_type)

            if entity_type_handle is None:
                return

            if "Color map" in entity_type_handle:
                del entity_type_handle["Color map"]
                entity_type.workspace.repack = True

            if color_map is not None and color_map.values is not None:
                H5Writer.create_dataset(
                    entity_type_handle,
                    color_map._values,  # pylint: disable=protected-access
                    "Color map",
                )
                entity_type_handle["Color map"].attrs.create(
                    "File name", color_map.name, dtype=H5Writer.str_type
                )

    @staticmethod
    def write_value_map(
        file: str | h5py.File,
        entity_type: ReferenceDataType,
        name="Value map",
        value_map: ReferenceValueMap | None = None,
    ) -> None:
        """
        Add :obj:`~geoh5py.data.reference_value_map.ReferenceValueMap` to a
        :obj:`~geoh5py.data.data_type.DataType`.

        :param file: Name or handle to a geoh5 file
        :param entity_type: Target entity_type with value_map
        :param name: Name of the value map
        :param value_map: Value map to be written
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            if isinstance(entity_type, GeometricDataValueMapType):
                return

            entity_type_handle = H5Writer.fetch_handle(h5file, entity_type)

            if entity_type_handle is None:
                return

            dtype = np.dtype([("Key", "<u4"), ("Value", H5Writer.str_type)])

            if value_map is None:
                value_map = entity_type.value_map

            if value_map is None:
                return

            if not isinstance(value_map, ReferenceValueMap):
                raise TypeError("Value map must be a ReferenceValueMap object.")

            if name in entity_type_handle:
                del entity_type_handle[name]
                entity_type.workspace.repack = True

            H5Writer.create_dataset(
                entity_type_handle,
                value_map.map.astype(dtype),
                name,
            )

            if name != "Value map":
                entity_type_handle[name].attrs.create(
                    "Name", value_map.name, dtype=H5Writer.str_type
                )
                entity_type_handle[name].attrs.create(
                    "Allow delete", True, dtype="int8"
                )

    @staticmethod
    def write_data_map(file: str | h5py.File, data: ReferencedData):
        """
        Write the value map of geometric data.

        :param file: Name or handle to a geoh5 file
        :param data: Target referenced data with value map
        """
        if data.data_maps is None:
            return

        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_type_handle = H5Writer.fetch_handle(h5file, data.entity_type)

            if entity_type_handle is None:
                return

            for name in entity_type_handle:
                if re.match("Value map [0-9]", name):
                    del entity_type_handle[name]

            for ii, child in enumerate(data.data_maps.values()):
                H5Writer.write_value_map(
                    h5file,
                    data.entity_type,
                    f"Value map {ii + 1}",
                    child.entity_type.value_map,
                )

    @staticmethod
    def write_visible(
        file: str | h5py.File,
        entity,
    ) -> None:
        """
        Needs revision once Visualization is implemented

        :param file: Name or handle to a geoh5 file
        :param entity: Target entity
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle = H5Writer.fetch_handle(h5file, entity)

            if entity_handle is None:
                return

            dtype = np.dtype(
                [("ViewID", h5py.special_dtype(vlen=str)), ("Visible", "int8")]
            )

            if entity.visible:
                visible = entity_handle.create_dataset(
                    "Visible", shape=(1,), dtype=dtype
                )
                visible["Visible"] = 1

    @staticmethod
    def write_array_attribute(
        file: str | h5py.File, entity, attribute, values=None, **kwargs
    ) -> None:
        """
        Add :obj:`~geoh5py.objects.object_base.ObjectBase.surveys` of an object.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target entity.
        :param attribute: Name of the attribute to be written to geoh5
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle = H5Writer.fetch_handle(h5file, entity)

            if entity_handle is None:
                return

            if values is None and getattr(entity, f"{attribute}", None) is not None:
                values = getattr(entity, f"_{attribute}", None)

            if (
                isinstance(entity, Concatenator)
                and attribute != "concatenated_object_ids"
            ):
                entity_handle = entity_handle["Concatenated Data"]

            if KEY_MAP[attribute] in entity_handle:
                del entity_handle[KEY_MAP[attribute]]
                entity.workspace.repack = True

            if isinstance(values, np.ndarray) and np.issubdtype(values.dtype, np.str_):
                values = values.astype(h5py.special_dtype(vlen=str))

            if values is not None:
                entity_handle.create_dataset(
                    KEY_MAP[attribute],
                    data=values,
                    compression="gzip",
                    compression_opts=9,
                    **kwargs,
                )

    @staticmethod
    def prepare_data_to_write(
        h5file: h5py.File, entity: Data | Group | ObjectBase, attribute: str
    ) -> tuple[h5py.Group, str] | tuple[None, None]:
        """
        Prepare data to be written to a geoh5 file.

        :param h5file: Name or handle to a geoh5 file.
        :param entity: Entity with attributes to be added to the geoh5 file.
        :param attribute: The attribute to be written to the geoh5 file.

        :return: The entity handle, the name map, and the values to be written to the geoh5 file.
        """
        entity_handle = H5Writer.fetch_handle(h5file, entity)

        if entity_handle is None:
            return None, None

        name_map = KEY_MAP[attribute]

        if isinstance(entity, Concatenator) and name_map != "Metadata":
            entity_handle = entity_handle["Concatenated Data"]

            if (
                attribute == "concatenated_attributes"
                and entity.concat_attr_str == "Attributes Jsons"
            ):
                name_map = entity.concat_attr_str

        if name_map in entity_handle:
            del entity_handle[name_map]
            entity.workspace.repack = True

        return entity_handle, name_map

    @staticmethod
    def write_group_values(
        file: str | h5py.File,
        entity: Group,
        attribute,
        compression: int,
        values=None,
        **kwargs,
    ):
        """
        Add Concatenator values.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target entity.
        :param attribute: Name of the attribute to be written to geoh5
        :param compression: Compression level for the data.
        :param values: Data values.
        """
        if not isinstance(entity, Group):
            raise TypeError(
                "Entity must be a Group object; " f"got '{type(entity)}' instead."
            )
        # check type here
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle, name_map = H5Writer.prepare_data_to_write(
                h5file, entity, attribute
            )

            # Give the chance to fetch from file
            values = getattr(entity, attribute, None) if values is None else values

            if values is None or entity_handle is None:
                return

            if (
                attribute == "concatenated_attributes"
                and isinstance(entity, Concatenator)
                and entity.concat_attr_str == "Attributes Jsons"
            ):
                values = [
                    json.dumps(val).encode("utf-8") for val in values["Attributes"]
                ]

            if name_map == "Attributes Jsons":
                entity_handle.create_dataset(
                    name_map,
                    data=values,
                    compression="gzip",
                    compression_opts=compression,
                    **kwargs,
                )
            elif isinstance(values, dict):
                values = deepcopy(values)
                values = dict_mapper(values, [as_str_if_uuid])
                entity_handle.create_dataset(
                    name_map,
                    data=json.dumps(values, indent=4),
                    dtype=h5py.special_dtype(vlen=str),
                    shape=(1,),
                    **kwargs,
                )
            else:
                warn(f"Writing '{values}' on '{entity.name}' failed.")

    @staticmethod
    def write_data_values(  # pylint: disable=too-many-branches
        file: str | h5py.File,
        entity: Data,
        attribute,
        compression: int,
        **kwargs,
    ) -> None:
        """
        Add data :obj:`~geoh5py.data.data.Data.values`.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target entity.
        :param attribute: Name of the attribute to be written to geoh5
        :param compression: Compression level for the data.
        """
        if not isinstance(entity, Data):
            raise TypeError(
                "Entity must be a Data object; " f"got '{type(entity)}' instead."
            )

        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle, name_map = H5Writer.prepare_data_to_write(
                h5file, entity, attribute
            )
            if entity.formatted_values is None or entity_handle is None:
                return
            if isinstance(entity, CommentsData):
                entity_handle.create_dataset(
                    name_map,
                    data=entity.formatted_values,
                    dtype=h5py.special_dtype(vlen=str),
                    shape=(1,),
                    **kwargs,
                )
            elif isinstance(entity, FilenameData):
                H5Writer.write_file_name_data(entity_handle, entity)
            elif isinstance(entity.values, str):
                entity_handle.create_dataset(
                    name_map,
                    data=entity.values,
                    dtype=h5py.special_dtype(vlen=str),
                    shape=(1,),
                    **kwargs,
                )
            else:
                entity_handle.create_dataset(
                    name_map,
                    data=entity.formatted_values,
                    compression="gzip",
                    compression_opts=compression,
                    **kwargs,
                )

    @staticmethod
    def write_metadata(  # pylint: disable=too-many-branches
        file: str | h5py.File,
        entity: Group | ObjectBase,
        attribute,
        values=None,
        **kwargs,
    ) -> None:
        """
        Add data :obj:`~geoh5py.data.data.Data.values`.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target entity.
        :param attribute: Name of the attribute to be written to geoh5
        :param values: Data values.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle, name_map = H5Writer.prepare_data_to_write(
                h5file, entity, attribute
            )

            # Give the chance to fetch from file
            values = getattr(entity, attribute, None) if values is None else values

            if values is None or entity_handle is None:
                return

            values = deepcopy(values)
            values = dict_mapper(values, [as_str_if_uuid])
            entity_handle.create_dataset(
                name_map,
                data=json.dumps(values, indent=4),
                dtype=h5py.special_dtype(vlen=str),
                shape=(1,),
                **kwargs,
            )

    @staticmethod
    def clear_stats_cache(
        file: str | h5py.File,
        entity: Data,
    ) -> None:
        """
        Clear the StatsCache dataset.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target entity.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            if not isinstance(entity, Data):
                return

            entity_type_handle = H5Writer.fetch_handle(h5file, entity.entity_type)
            if entity_type_handle is None:
                return

            stats_cache = entity_type_handle.get("StatsCache")
            if stats_cache is not None:
                del entity_type_handle["StatsCache"]
                entity.workspace.repack = True

    @staticmethod
    def write_entity(
        file: str | h5py.File,
        entity,
        compression: int,
    ) -> h5py.Group:
        """
        Add an :obj:`~geoh5py.shared.entity.Entity` and its attributes to geoh5.
        The function returns a pointer to the entity if already present on file.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target :obj:`~geoh5py.shared.entity.Entity`.
        :param compression: Compression level for data.

        :return entity: Pointer to the written entity. Active link if "close_file" is False.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            base = list(h5file)[0]
            entity_type = str_from_type(entity)
            uid = entity.uid

            if entity_type not in h5file[base]:
                h5file[base].create_group(entity_type)

            # Check if already in the project
            if as_str_if_uuid(uid) in h5file[base][entity_type]:
                entity.on_file = True

                return h5file[base][entity_type][as_str_if_uuid(uid)]

            entity_handle = h5file[base][entity_type].create_group(as_str_if_uuid(uid))
            if isinstance(entity, Concatenator):
                concat_group = entity_handle.create_group("Concatenated Data")
                entity_handle.create_group("Data")
                concat_group.create_group("Index")
                concat_group.create_group("Data")
                entity_handle.create_group("Groups")
            elif entity_type == "Groups":
                entity_handle.create_group("Data")
                entity_handle.create_group("Groups")
                entity_handle.create_group("Objects")
            elif entity_type == "Objects":
                entity_handle.create_group("Data")

            # Add the type
            new_type = H5Writer.write_entity_type(h5file, entity.entity_type)
            entity_handle["Type"] = new_type
            entity.entity_type.on_file = True

            H5Writer.write_properties(h5file, entity, compression)
            entity.on_file = True

            if isinstance(entity, RootGroup):
                if "Root" in h5file[base]:
                    del h5file[base]["Root"]

                h5file[base]["Root"] = entity_handle

        return entity_handle

    @staticmethod
    def write_entity_type(
        file: str | h5py.File,
        entity_type: shared.EntityType | ReferenceDataType,
    ) -> h5py.Group:
        """
        Add an :obj:`~geoh5py.shared.entity_type.EntityType` to geoh5.

        :param file: Name or handle to a geoh5 file.
        :param entity_type: Entity with type to be added.

        :return type: Pointer to :obj:`~geoh5py.shared.entity_type.EntityType` in geoh5.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            base = list(h5file)[0]
            uid = entity_type.uid
            entity_type_str = str_from_subtype(entity_type)

            if "Types" not in h5file[base]:
                h5file[base].create_group("Types")

            # Check if already in the project
            if entity_type_str not in h5file[base]["Types"]:
                h5file[base]["Types"].create_group(entity_type_str)

            if as_str_if_uuid(uid) in h5file[base]["Types"][entity_type_str]:
                entity_type.on_file = True

                return h5file[base]["Types"][entity_type_str][as_str_if_uuid(uid)]

            new_type = h5file[base]["Types"][entity_type_str].create_group(
                as_str_if_uuid(uid)
            )
            H5Writer.write_attributes(h5file, entity_type)

            if hasattr(entity_type, "color_map"):
                H5Writer.write_color_map(h5file, entity_type)

            if isinstance(entity_type, ReferenceDataType):
                H5Writer.write_value_map(h5file, entity_type)

            entity_type.on_file = True

        return new_type

    @staticmethod
    def write_file_name_data(entity_handle: h5py.Group, entity: FilenameData) -> None:
        """
        Write a dataset for the file name and file blob.

        :param entity_handle: Pointer to the geoh5 Group.
        :param entity: Target :obj:`~geoh5py.data.filename_data.FilenameData` entity.
        :param values: Bytes data
        """
        if entity.file_bytes is None or entity.values is None:
            raise AttributeError("FilenameData requires the 'file_bytes' to be set.")

        entity_handle.create_dataset(
            "Data",
            data=entity.values,
            dtype=h5py.special_dtype(vlen=str),
            shape=(1,),
        )

        if entity.values in entity_handle:
            del entity_handle[entity.values]
            entity.workspace.repack = True

        entity_handle.create_dataset(
            entity.values,
            data=np.asarray(np.void(entity.file_bytes[:])),
            shape=(1,),
        )

    @staticmethod
    def write_properties(
        file: str | h5py.File,
        entity: Entity,
        compression: int,
    ) -> None:
        """
        Add properties of an :obj:`~geoh5py.shared.entity.Entity`.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target :obj:`~geoh5py.shared.entity.Entity`.
        :param compression: Compression level for data.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            H5Writer.update_field(h5file, entity, "attributes", compression)

            for attribute in KEY_MAP:
                if getattr(entity, attribute, None) is not None:
                    H5Writer.update_field(h5file, entity, attribute, compression)

    @staticmethod
    def write_property_groups(
        file: str | h5py.File,
        entity,
    ) -> None:
        """
        Write :obj:`~geoh5py.groups.property_group.PropertyGroup` associated with
        an :obj:`~geoh5py.shared.entity.Entity`.

        :param file: Name or handle to a geoh5 file.
        :param entity: Target :obj:`~geoh5py.shared.entity.Entity`.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle = H5Writer.fetch_handle(h5file, entity)
            if entity_handle is None:
                return

            if "PropertyGroups" in entity_handle:
                del entity_handle["PropertyGroups"]
                entity.workspace.repack = True

            if hasattr(entity, "property_groups") and isinstance(
                entity.property_groups, list
            ):
                for p_g in entity.property_groups:
                    H5Writer.add_or_update_property_group(h5file, p_g)

    @staticmethod
    def add_or_update_property_group(file, property_group, remove=False):
        """
        Update a :obj:`~geoh5py.groups.property_group.PropertyGroup` associated with
        an :obj:`~geoh5py.shared.entity.Entity`.

        :param file: Name or handle to a geoh5 file.
        :param property_group: Target PropertyGroup
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            entity_handle = H5Writer.fetch_handle(h5file, property_group.parent)

            if entity_handle is None:
                return

            if "PropertyGroups" not in entity_handle:
                entity_handle.create_group("PropertyGroups")

            uid = as_str_if_uuid(property_group.uid)
            if uid in entity_handle["PropertyGroups"]:
                del entity_handle["PropertyGroups"][uid]
                property_group.parent.workspace.repack = True

            if remove:
                return

            entity_handle["PropertyGroups"].create_group(uid)
            group_handle = entity_handle["PropertyGroups"][uid]

            for key, attr in property_group.attribute_map.items():
                try:
                    value = getattr(property_group, attr)
                except AttributeError:
                    continue

                if key == "Association":
                    value = value.name.capitalize()

                elif key == "Properties":
                    if value is None:
                        continue

                    value = np.asarray([as_str_if_uuid(val) for val in value])

                elif key == "ID":
                    value = as_str_if_uuid(value)
                elif key == "Property Group Type":
                    value = value.value

                group_handle.attrs.create(
                    key, value, dtype=h5py.special_dtype(vlen=str)
                )

            property_group.on_file = True

    @staticmethod
    def write_to_parent(
        file: str | h5py.File,
        entity: Entity,
        compression: int,
        recursively: bool = False,
    ) -> None:
        """
        Add/create an :obj:`~geoh5py.shared.entity.Entity` and add it to its parent.

        :param file: Name or handle to a geoh5 file.
        :param entity: Entity to be added or linked to a parent in geoh5.
        :param compression: Compression level for data.
        :param recursively: Add parents recursively until reaching the
            :obj:`~geoh5py.groups.root_group.RootGroup`.
        """
        with fetch_h5_handle(file, mode="r+") as h5file:
            if isinstance(entity, RootGroup):
                return

            uid = entity.uid
            entity_handle = H5Writer.write_entity(h5file, entity, compression)
            parent_handle = H5Writer.write_entity(h5file, entity.parent, compression)
            entity_type = str_from_type(entity)

            # Check if child h5py.Group already exists
            if entity_type not in parent_handle:
                parent_handle.create_group(entity_type)

            # Check if child uuid not already in h5
            if as_str_if_uuid(uid) not in parent_handle[entity_type]:
                parent_handle[entity_type][as_str_if_uuid(uid)] = entity_handle

            if recursively:
                H5Writer.write_to_parent(
                    h5file, entity.parent, compression=compression, recursively=True
                )



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING
from uuid import UUID, uuid4
from warnings import warn

from ..data import Data, DataAssociationEnum
from ..shared.utils import (
    find_unique_name,
    remove_duplicates_in_list,
)
from .property_group_table import PropertyGroupTable
from .property_group_type import GroupTypeEnum


if TYPE_CHECKING:  # pragma: no cover
    from ..objects import ObjectBase


class PropertyGroup:
    """
    Property group listing data children of an object.
    This group is not registered to the workspace and only visible to the parent object.

    :param parent: Parent object.
    :param association: Association of the data.
    :param allow_delete: Allow deleting the group.
    :param name: Name of the group.
    :param on_file: Property group is on file.
    :param uid: Unique identifier.
    :param property_group_type: Type of property group.
    :param properties: List of data or unique identifiers for the data.
    """

    _attribute_map = {
        "Association": "association",
        "Group Name": "name",
        "ID": "uid",
        "Properties": "properties",
        "Property Group Type": "property_group_type",
    }

    def __init__(  # pylint: disable=too-many-arguments
        self,
        parent: ObjectBase,
        *,
        association: str | DataAssociationEnum | None = None,
        allow_delete: bool = True,
        name: str = "Property Group",
        on_file: bool = False,
        uid: UUID | None = None,
        property_group_type: GroupTypeEnum | str = GroupTypeEnum.SIMPLE,
        properties: list[UUID | Data | str] | None = None,
        **_,
    ):
        self._parent: ObjectBase = self._validate_parent(parent)
        self._property_group_type = self._validate_group_type(property_group_type)

        self.allow_delete = allow_delete
        self.name = name
        self.on_file = on_file
        self.uid = uid or uuid4()

        properties_list = self._initialize_properties(properties)

        self._association = self._validate_association(association, properties_list)
        self._properties = self._validate_properties(properties_list)

        self.parent.add_children([self])
        self.parent.workspace.register(self)

    def _initialize_properties(
        self, properties: str | UUID | Data | list[UUID | str | Data] | None
    ) -> list[Data] | None:
        """
        Initialize the properties list.

        :param properties: List of Data entities to validate.

        :return: List of unique identifiers for the Data entities.
        """
        if properties is None:
            return None

        if not isinstance(properties, Iterable):
            properties = [properties]

        return [self.parent.reference_to_data(prop) for prop in properties]

    @staticmethod
    def _validate_association(
        value: str | DataAssociationEnum | None, properties: list[Data] | None
    ) -> DataAssociationEnum:
        """
        Verify that the association is valid, or infer it from the properties.

        :param value: The association to validate.
        :param properties: A list of properties to infer the association from.
        """
        if properties is None and value is None:
            raise ValueError(
                "At least one of 'properties' or 'association' must be provided."
            )

        if value is None and properties is not None:
            value = properties[0].association

        if isinstance(value, str):
            value = getattr(DataAssociationEnum, value.upper())

        if not isinstance(value, DataAssociationEnum):
            raise TypeError(f"Association must be one of type {DataAssociationEnum}")

        return value

    def _validate_data(self, data: Data | UUID | str) -> Data:
        """
        Verify that the data is in the parent and has the same association as the group.

        :param data: The data to verify.
            It can be the name, the uuid or the data itself.

        :return: The uuid of the data.
        """
        data = self.parent.reference_to_data(data)

        if self.association != data.association:
            raise ValueError(
                f"Data '{data.name}' association ({data.association}) "
                f"does not match group association ({self.association})."
            )

        return data

    @staticmethod
    def _validate_group_type(value: str | GroupTypeEnum) -> GroupTypeEnum:
        """
        Verify that the group type is a valid GroupTypeEnum.

        :param value: The group type to validate.

        :return: The validated group type.
        """
        if isinstance(value, str):
            try:
                value = GroupTypeEnum(value)
            except ValueError as error:
                raise ValueError(
                    f"'Property group type' must be one of "
                    f"{', '.join(GroupTypeEnum.__members__)}. Provided {value}"
                ) from error

        if not isinstance(value, GroupTypeEnum):
            raise TypeError(
                f"'Property group type' must be of type {GroupTypeEnum}, "
                f"provided {type(value)}"
            )

        return value

    @staticmethod
    def _validate_parent(parent: ObjectBase) -> ObjectBase:
        """
        Verify that the parent is valid.

        :param parent: The parent Object to validate.

        :return: The parent Object.
        """
        # define the parent
        if not hasattr(parent, "_property_groups"):
            raise TypeError(f"Parent {parent} must have a 'property_groups' attribute")
        return parent

    def _validate_properties(
        self, data_list: Sequence[str | UUID | Data] | None
    ) -> list[UUID] | None:
        """
        Validate the properties list.

        :param data_list: List of Data entities to validate.

        :return: List of unique identifiers for the Data entities.
        """
        if not data_list:
            return None

        data_list_ = remove_duplicates_in_list(
            [self._validate_data(uid) for uid in data_list]
        )

        self._property_group_type.verify(data_list_)

        return [data.uid for data in data_list_]

    def add_properties(self, data: str | Data | list[str | Data | UUID] | UUID):
        """
        Add data to properties.

        :param data: Data to add to the group.
            It can be the name, the uuid or the data itself in a list or alone.
        """
        if self._property_group_type.no_modify:
            raise ValueError(
                f"Cannot add properties to '{self._property_group_type}' property group type."
            )

        if not isinstance(data, list):
            data = [data]

        properties = self._validate_properties(
            data if self.properties is None else self.properties + data
        )

        if properties:
            self._properties = remove_duplicates_in_list(properties)
            self.parent.workspace.add_or_update_property_group(self)

    @property
    def allow_delete(self) -> bool:
        """
        Allow deleting the group
        """
        return self._allow_delete

    @allow_delete.setter
    def allow_delete(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("allow_delete must be a boolean")
        self._allow_delete = value

    @property
    def association(self) -> DataAssociationEnum:
        """
        The association of the data.
        """
        return self._association

    @property
    def attribute_map(self) -> dict:
        """
        Attribute names mapping between geoh5 and geoh5py
        """
        return self._attribute_map

    @property
    def collect_values(self) -> list | None:
        """
        The values of the properties in the group.
        """
        warn(
            "PropertyGroup.collect_values is deprecated, use PropertyGroup.table instead.",
            DeprecationWarning,
        )

        if self._properties is None:
            return None

        return [self._parent.get_data(data)[0].values for data in self._properties]

    @property
    def name(self) -> str:
        """
        Name of the group
        """
        return self._name

    @name.setter
    def name(self, new_name: str):
        if not isinstance(new_name, str):
            raise TypeError("Name must be a string")

        self._name = new_name

    @property
    def on_file(self):
        """
        Property group is on geoh5 file.
        """
        return self._on_file

    @on_file.setter
    def on_file(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("Attribute 'on_file' must be a boolean.")

        self._on_file = value

    @property
    def parent(self) -> ObjectBase:
        """
        The parent of the PropertyGroup.
        """
        return self._parent

    @property
    def properties(self) -> list[UUID] | None:
        """
        List of unique identifiers for the :obj:`~geoh5py.data.data.Data`
        contained in the property group.
        """
        return self._properties

    @property
    def properties_name(self) -> list[str] | None:
        """
        List of names of the properties`
        """
        if self._properties is None:
            return None

        names: list[str] = []
        for uid in self._properties:
            data = self.parent.get_data(uid)[0]
            name = str(data.uid) if data.name is None else data.name
            names.append(find_unique_name(name, names))

        return names

    @property
    def property_group_type(self) -> GroupTypeEnum:
        """
        Type of property group.
        """
        return self._property_group_type

    def remove_properties(self, data: str | Data | list[str | Data | UUID] | UUID):
        """
        Remove data from the properties.

        :param data: Data to remove from the group.
            It can be the name, the uuid or the data itself in a list or alone.
        """
        if self._property_group_type.no_modify:
            raise ValueError(
                f"Cannot remove properties from '{self._property_group_type}' property group type."
            )

        if self.properties is None:
            return

        if not isinstance(data, list):
            data = [data]

        properties = self.properties
        for elem in data:
            elem = self.parent.reference_to_data(elem).uid
            if elem in properties:
                properties.remove(elem)

        self._properties = self._validate_properties(properties)

        if not self._properties:
            self.parent.workspace.remove_entity(self)
            return

        self.parent.workspace.add_or_update_property_group(self)

    @property
    def table(self) -> PropertyGroupTable:
        """
        Create an object to access the data of the property group.
        """
        return PropertyGroupTable(self)

    @property
    def uid(self) -> UUID:
        """
        Unique identifier
        """
        return self._uid

    @uid.setter
    def uid(self, uid: str | UUID):
        if isinstance(uid, str):
            uid = UUID(uid)

        if not isinstance(uid, UUID):
            raise TypeError(f"Could not convert input uid {uid} to type UUID")

        self._uid = uid



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

from .object_base import ObjectBase


class Slicer(ObjectBase):
    """
    Slicer object.

    It's an empty object used by
    :class:`geoh5py.groups.interpretation_section.InterpretationSection`
    to store the slicing VisualParameter information.
    """

    _attribute_map: dict = ObjectBase._attribute_map.copy()
    _attribute_map.pop("Clipping IDs")
    _attribute_map.pop("PropertyGroups")
    _default_name = "Slicer"
    _TYPE_UID = uuid.UUID("{238F961D-AE63-43DE-AB64-E1A079271CF5}")



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from abc import ABC
from collections.abc import Callable
from contextlib import contextmanager
from io import BytesIO
from json import loads
from pathlib import Path
from typing import TYPE_CHECKING, Any
from uuid import UUID
from warnings import warn

import h5py
import numpy as np

from .exceptions import Geoh5FileClosedError


if TYPE_CHECKING:
    from ..workspace import Workspace
    from .entity import Entity

INV_KEY_MAP = {
    "Allow delete": "allow_delete",
    "Allow delete contents": "allow_delete_content",
    "Allow move": "allow_move",
    "Allow move contents": "allow_move_content",
    "Allow rename": "allow_rename",
    "Association": "association",
    "Attributes": "concatenated_attributes",
    "Blob": "BLOB",
    "Boolean": "BOOLEAN",
    "Cell": "CELL",
    "Cells": "cells",
    "Clipping IDs": "clipping_ids: list | None",
    "Collar": "collar",
    "Color map": "color_map",
    "Contributors": "contributors",
    "Concatenated object IDs": "concatenated_object_ids",
    "Cost": "cost",
    "Current line property ID": "current_line_id",
    "Data": "values",
    "DateTime": "DATETIME",
    "Description": "description",
    "Dip": "dip",
    "Distance unit": "distance_unit",
    "Dynamic implementation ID": "dynamic_implementation_id",
    "Duplicate type on copy": "duplicate_type_on_copy",
    "End of hole": "end_of_hole",
    "Face": "FACE",
    "File name": "name",
    "Filename": "FILENAME",
    "Float": "FLOAT",
    "Geometric": "GEOMETRIC",
    "Group": "GROUP",
    "Group Name": "name",
    "GA Version": "ga_version",
    "Hidden": "hidden",
    "Invalid": "INVALID",
    "Integer": "INTEGER",
    "ID": "uid",
    "Last focus": "last_focus",
    "Layers": "layers",
    "Mapping": "mapping",
    "Metadata": "metadata",
    "Modifiable": "modifiable",
    "Multi-Text": "MULTI_TEXT",
    "Name": "name",
    "Number of bins": "number_of_bins",
    "NU": "u_count",
    "NV": "v_count",
    "NW": "w_count",
    "options": "options",
    "Object": "OBJECT",
    "Origin": "origin",
    "Octree Cells": "octree_cells",
    "Partially hidden": "partially_hidden",
    "Planning": "planning",
    "Precision": "precision",
    "Primitive type": "primitive_type",
    "Prisms": "prisms",
    "Properties": "properties",
    "Property Group IDs": "property_group_ids",
    "Property Group Type": "property_group_type",
    "PropertyGroups": "property_groups",
    "Public": "public",
    "Referenced": "REFERENCED",
    "Rotation": "rotation",
    "Scale": "scale",
    "Scientific notation": "scientific_notation",
    "Surveys": "surveys",
    "Text": "TEXT",
    "Trace": "trace",
    "TraceDepth": "trace_depth",
    "Transparent no data": "transparent_no_data",
    "Unknown": "UNKNOWN",
    "U cell delimiters": "u_cell_delimiters",
    "V cell delimiters": "v_cell_delimiters",
    "Z cell delimiters": "z_cell_delimiters",
    "U Cell Size": "u_cell_size",
    "U Count": "u_count",
    "U Size": "u_cell_size",
    "V Cell Size": "v_cell_size",
    "V Count": "v_count",
    "V Size": "v_cell_size",
    "Vector": "VECTOR",
    "Version": "version",
    "Vertical": "vertical",
    "Vertices": "vertices",
    "Vertex": "VERTEX",
    "Visible": "visible",
    "W Cell Size": "w_cell_size",
}

KEY_MAP = {value: key for key, value in INV_KEY_MAP.items()}

PNG_KWARGS = {"format": "PNG", "compress_level": 9}
JPG_KWARGS = {"format": "JPEG", "quality": 85}
TIF_KWARGS = {"format": "TIFF"}

PILLOW_ARGUMENTS = {
    "1": PNG_KWARGS,
    "L": PNG_KWARGS,
    "P": PNG_KWARGS,
    "RGB": PNG_KWARGS,
    "RGBA": PNG_KWARGS,
    "CMYK": JPG_KWARGS,
    "YCbCr": JPG_KWARGS,
    "I": TIF_KWARGS,
    "F": TIF_KWARGS,
}


@contextmanager
def fetch_active_workspace(workspace: Workspace | None, mode: str = "r"):
    """
    Open a workspace in the requested 'mode'.

    If receiving an opened Workspace instead, merely return the given workspace.

    :param workspace: A Workspace class
    :param mode: Set the h5 read/write mode

    :return h5py.File: Handle to an opened Workspace.
    """
    try:
        geoh5 = None if workspace is None else workspace.geoh5
    except Geoh5FileClosedError:
        geoh5 = None

    if workspace is None or (geoh5 is not None and mode in workspace.geoh5.mode):
        try:
            yield workspace
        finally:
            pass
    else:
        if geoh5 is not None:
            warn(
                f"Closing the workspace in mode '{workspace.geoh5.mode}' "
                f"and re-opening in mode '{mode}'."
            )
            workspace.close()

        try:
            yield workspace.open(mode=mode)
        finally:
            workspace.close()


@contextmanager
def fetch_h5_handle(file: str | h5py.File | Path, mode: str = "r") -> h5py.File:
    """
    Open in read+ mode a geoh5 file from string.
    If receiving a file instead of a string, merely return the given file.

    :param file: Name or handle to a geoh5 file.
    :param mode: Set the h5 read/write mode

    :return h5py.File: Handle to an opened h5py file.
    """
    if isinstance(file, h5py.File):
        try:
            yield file
        finally:
            pass
    else:
        if Path(file).suffix != ".geoh5":
            raise ValueError("Input h5 file must have a 'geoh5' extension.")

        h5file = h5py.File(file, mode)

        try:
            yield h5file
        finally:
            h5file.close()


def match_values(vec_a, vec_b, collocation_distance=1e-4) -> np.ndarray:
    """
    Find indices of matching values between two arrays, within collocation_distance.

    :param: vec_a, list or numpy.ndarray
        Input sorted values

    :param: vec_b, list or numpy.ndarray
        Query values

    :return: indices, numpy.ndarray
        Pairs of indices for matching values between the two arrays such
        that vec_a[ind[:, 0]] == vec_b[ind[:, 1]].
    """
    ind_sort = np.argsort(vec_a)
    ind = np.minimum(
        np.searchsorted(vec_a[ind_sort], vec_b, side="right"), vec_a.shape[0] - 1
    )
    nearests = np.c_[ind, ind - 1]
    match = np.where(
        np.abs(vec_a[ind_sort][nearests] - vec_b[:, None]) < collocation_distance
    )
    indices = np.c_[ind_sort[nearests[match[0], match[1]]], match[0]]
    return indices


def merge_arrays(
    head,
    tail,
    *,
    replace="A->B",
    mapping=None,
    collocation_distance=1e-4,
    return_mapping=False,
) -> np.ndarray:
    """
    Given two numpy.arrays of different length, find the matching values and append both arrays.

    :param: head, numpy.array of float
        First vector of shape(M,) to be appended.
    :param: tail, numpy.array of float
        Second vector of shape(N,) to be appended
    :param: mapping=None, numpy.ndarray of int
        Optional array where values from the head are replaced by the tail.
    :param: collocation_distance=1e-4, float
        Tolerance between matching values.

    :return: numpy.array shape(O,)
        Unique values from head to tail without repeats, within collocation_distance.
    """

    if mapping is None:
        mapping = match_values(head, tail, collocation_distance=collocation_distance)

    if mapping.shape[0] > 0:
        if replace == "B->A":
            head[mapping[:, 0]] = tail[mapping[:, 1]]
        else:
            tail[mapping[:, 1]] = head[mapping[:, 0]]

        tail = np.delete(tail, mapping[:, 1])

    if return_mapping:
        return np.r_[head, tail], mapping

    return np.r_[head, tail]


def clear_array_attributes(entity: Entity, recursive: bool = False):
    """
    Clear all stashed values of attributes from an entity to free up memory.

    :param entity: Entity to clear attributes from.
    :param recursive: Clear attributes from children entities.
    """
    if isinstance(entity.workspace.h5file, BytesIO):
        return

    for attribute in [
        "vertices",
        "cells",
        "values",
        "prisms",
        "layers",
        "octree_cells",
    ]:
        if hasattr(entity, attribute):
            setattr(entity, f"_{attribute}", None)

    if recursive and hasattr(entity, "children"):
        for child in entity.children:
            clear_array_attributes(child, recursive=recursive)


def are_objects_similar(obj1, obj2, ignore: list[str] | None):
    """
    Compare two objects to see if they are similar. This is a shallow comparison.

    :param obj1: The first object.
    :param obj2: The first object.
    :param ignore: List of attributes to ignore.

    :return: If attributes similar or not.
    """
    assert isinstance(obj1, type(obj2)), "Objects are not the same type."

    attributes1 = getattr(obj1, "__dict__", obj1)
    attributes2 = getattr(obj2, "__dict__", obj2)

    # remove the ignore attributes
    if isinstance(ignore, list) and isinstance(attributes1, dict):
        for item in ignore:
            attributes1.pop(item, None)
            attributes2.pop(item, None)

    return attributes1 == attributes2


def compare_arrays(object_a, object_b, attribute: str, decimal: int = 6):
    """
    Utility to compare array properties from two Entities

    :param object_a: First Entity
    :param object_b: Second Entity
    :param attribute: Attribute to compare
    :param decimal: Decimal precision for comparison
    """
    array_a_values = getattr(object_a, attribute)
    array_b_values = getattr(object_b, attribute)

    if array_b_values is None:
        raise ValueError(f"attr {attribute} is None for object {object_b.name}")

    if array_b_values.dtype.names is not None:
        assert all(
            np.all(array_a_values[name] == array_b_values[name])
            for name in array_b_values.dtype.names
        ), f"Error comparing attribute '{attribute}'."

    elif len(array_a_values) > 0 and isinstance(array_a_values[0], str):
        assert all(
            array_a_values == array_b_values
        ), f"Error comparing attribute '{attribute}'."
    else:
        np.testing.assert_array_almost_equal(
            array_a_values,
            array_b_values,
            decimal=decimal,
            err_msg=f"Error comparing attribute '{attribute}'.",
        )


def compare_floats(object_a, object_b, attribute: str, decimal: int = 6):
    np.testing.assert_almost_equal(
        getattr(object_a, attribute),
        getattr(object_b, attribute),
        decimal=decimal,
        err_msg=f"Error comparing attribute '{attribute}'.",
    )


def compare_list(object_a, object_b, attribute: str, ignore: list[str] | None):
    get_object_a = getattr(object_a, attribute)
    get_object_b = getattr(object_b, attribute)
    assert isinstance(get_object_a, list)
    assert len(get_object_a) == len(get_object_b)
    for obj_a, obj_b in zip(get_object_a, get_object_b, strict=False):
        assert are_objects_similar(obj_a, obj_b, ignore)


def compare_bytes(object_a, object_b):
    assert (
        object_a == object_b
    ), f"{type(object_a)} objects: {object_a}, {object_b} are not equal."


def compare_entities(
    object_a, object_b, ignore: list[str] | None = None, decimal: int = 6
) -> None:
    if isinstance(object_a, bytes):
        compare_bytes(object_a, object_b)
        return

    base_ignore = [
        "_workspace",
        "_children",
        "_visual_parameters",
        "_entity_class",
        "_intervals",
    ]
    ignore_list = base_ignore + ignore if ignore else base_ignore

    for attr in [k for k in object_a.__dict__ if k not in ignore_list]:
        if isinstance(getattr(object_a, attr.lstrip("_")), ABC):
            compare_entities(
                getattr(object_a, attr.lstrip("_")),
                getattr(object_b, attr.lstrip("_")),
                ignore=ignore,
                decimal=decimal,
            )
        else:
            if isinstance(getattr(object_a, attr.lstrip("_")), np.ndarray):
                compare_arrays(object_a, object_b, attr.lstrip("_"), decimal=decimal)
            elif isinstance(getattr(object_a, attr.lstrip("_")), float):
                compare_floats(object_a, object_b, attr.lstrip("_"), decimal=decimal)
            elif isinstance(getattr(object_a, attr.lstrip("_")), list):
                compare_list(object_a, object_b, attr.lstrip("_"), ignore)
            else:
                try:
                    assert np.all(
                        getattr(object_a, attr.lstrip("_"))
                        == getattr(object_b, attr.lstrip("_"))
                    ), (
                        f"Output attribute '{attr.lstrip('_')}' for {object_a} do "
                        f"not match input {object_b}"
                    )
                except AssertionError:
                    pass


def is_uuid(value: str) -> bool:
    """Check if a string is UUID compliant."""
    try:
        UUID(str(value))
        return True
    except ValueError:
        return False


def entity2uuid(value: Any) -> UUID | Any:
    """Convert an entity to its UUID."""
    if hasattr(value, "uid"):
        return value.uid
    return value


def uuid2entity(value: UUID, workspace: Workspace) -> Entity | Any:
    """Convert UUID to a known entity."""
    if isinstance(value, UUID):
        if value in workspace.list_entities_name:
            return workspace.get_entity(value)[0]

        # Search for property groups
        for obj in workspace.objects:
            if obj.property_groups is not None:
                prop_group = [
                    prop_group
                    for prop_group in obj.property_groups
                    if prop_group.uid == value
                ]

                if prop_group:
                    return prop_group[0]

        return None

    return value


def str2uuid(value: Any) -> UUID | Any:
    """Convert string to UUID"""
    if isinstance(value, bytes):
        value = value.decode("utf-8")

    if is_uuid(value):
        # TODO insert validation
        return UUID(str(value))
    return value


def as_str_if_uuid(value: UUID | Any) -> str | Any:
    """Convert :obj:`UUID` to string used in geoh5."""
    if isinstance(value, UUID):
        return "{" + str(value) + "}"
    return value


def bool_value(value: np.int8) -> bool:
    """Convert logical int8 to bool."""
    return bool(value)


def as_str_if_utf8_bytes(value) -> str:
    """Convert bytes to string"""
    if isinstance(value, bytes):
        value = value.decode("utf-8")
    return value


def as_float_if_isnumeric(value: str) -> float | str:
    """Convert bytes to string"""
    if value.isnumeric():
        return float(value)
    return value


def str_json_to_dict(string: str | bytes) -> dict:
    """
    Convert a json string or bytes to a dictionary.

    :param string: The json string or bytes to convert to a dictionary.

    :return: The dictionary representation of the json string with uuid promoted.
    """
    value = as_str_if_utf8_bytes(string)
    json_dict = loads(value)

    for key, val in json_dict.items():
        if isinstance(val, dict):
            for sub_key, sub_val in val.items():
                json_dict[key][sub_key] = str2uuid(sub_val)
        else:
            json_dict[key] = str2uuid(val)

    return json_dict


def ensure_uuid(value: UUID | str) -> UUID:
    """
    Ensure that the value is a UUID.

    If not, it raises a type error.

    :param value: The value to ensure is a UUID.

    :return: The verified UUID.
    """
    value = str2uuid(value)

    if not isinstance(value, UUID):
        raise TypeError(f"Value {value} is not a UUID but a {type(value)}.")

    return value


def dict_mapper(val, string_funcs: list[Callable], *args, omit: dict | None = None):
    """
    Recursion through nested dictionaries and applies mapping functions to values.

    :param val: Value (could be another dictionary) to apply transform functions.
    :param string_funcs: Functions to apply on values within the input dictionary.
    :param omit: Dictionary of functions to omit.

    :return val: Transformed values
    """
    if isinstance(val, dict):
        for key, values in val.items():
            short_list = string_funcs.copy()
            if omit is not None:
                short_list = [
                    fun for fun in string_funcs if fun not in omit.get(key, [])
                ]

            val[key] = dict_mapper(values, short_list)

    if isinstance(val, list):
        out = []
        for elem in val:
            for fun in string_funcs:
                elem = fun(elem, *args)
            out += [elem]
        return out

    for fun in string_funcs:
        val = fun(val, *args)
    return val


def box_intersect(extent_a: np.ndarray, extent_b: np.ndarray) -> bool:
    """
    Compute the intersection of two axis-aligned bounding extents defined by their
    arrays of minimum and maximum bounds in N-D space.

    :param extent_a: First extent or shape (2, N)
    :param extent_b: Second extent or shape (2, N)

    :return: Logic if the box extents intersect along all dimensions.
    """
    for extent in [extent_a, extent_b]:
        if not isinstance(extent, np.ndarray) or extent.ndim != 2:
            raise TypeError("Input extents must be 2D numpy.ndarrays.")

        if extent.shape[0] != 2 or not np.all(extent[0, :] <= extent[1, :]):
            raise ValueError(
                "Extents must be of shape (2, N) containing the minimum and maximum "
                "bounds in nd-space on the first and second row respectively."
            )

    for comp_a, comp_b in zip(extent_a.T, extent_b.T, strict=False):
        min_ext = max(comp_a[0], comp_b[0])
        max_ext = min(comp_a[1], comp_b[1])

        if min_ext > max_ext:
            return False

    return True


def mask_by_extent(
    locations: np.ndarray, extent: np.ndarray, inverse: bool = False
) -> np.ndarray:
    """
    Find indices of locations within a rectangular extent.

    :param locations: shape(*, 3) or shape(*, 2) Coordinates to be evaluated.
    :param extent: shape(2, 2) Limits defined by the South-West and
        North-East corners. Extents can also be provided as 3D coordinates
        with shape(2, 3) defining the top and bottom limits.
    :param inverse: Return the complement of the mask extent.

    :returns: Array of bool for the locations inside or outside the box extent.
    """
    if not isinstance(extent, np.ndarray) or extent.ndim != 2:
        raise ValueError("Input 'extent' must be a 2D array-like.")

    if not isinstance(locations, np.ndarray) or locations.ndim != 2:
        raise ValueError(
            "Input 'locations' must be an array-like of shape(*, 3) or (*, 2)."
        )

    indices = np.ones(locations.shape[0], dtype=bool)
    for loc, lim in zip(locations.T, extent.T, strict=False):
        indices &= (lim[0] <= loc) & (loc <= lim[1])

    if inverse:
        return ~indices

    return indices


def get_attributes(entity, omit_list=(), attributes=None) -> dict:
    """Extract the attributes of an object with omissions."""
    if attributes is None:
        attributes = {}
    for key in vars(entity):
        if key not in omit_list:
            key = key.lstrip("_")
            attr = getattr(entity, key)
            attributes[key] = attr

    return attributes


def dip_azimuth_to_vector(
    dip: float | np.ndarray, azimuth: float | np.ndarray
) -> np.ndarray:
    """
    Convert dip and azimuth to a unit vector.

    :param dip: The dip angle in degree from horizontal (positive up).
    :param azimuth: The azimuth angle in degree from North (clockwise).

    :return: The unit vector.
    """
    azimuth = np.deg2rad(90 - azimuth)
    dip = np.deg2rad(dip)

    return np.c_[
        np.cos(dip) * np.cos(azimuth),
        np.cos(dip) * np.sin(azimuth),
        np.sin(dip),
    ]


def xy_rotation_matrix(angle: float) -> np.ndarray:
    """
    Rotation matrix about the z-axis.

    :param angle: Rotation angle in radians.

    :return rot: Rotation matrix.
    """
    return np.array(
        [
            [np.cos(angle), -np.sin(angle), 0.0],
            [np.sin(angle), np.cos(angle), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )


def yz_rotation_matrix(angle: float) -> np.ndarray:
    """
    Rotation matrix about the x-axis.
    :param angle: Rotation angle in radians.
    :return: rot: Rotation matrix.
    """

    return np.array(
        [
            [1, 0, 0],
            [0, np.cos(angle), -np.sin(angle)],
            [0, np.sin(angle), np.cos(angle)],
        ]
    )


def dip_points(points: np.ndarray, dip: float, rotation: float = 0) -> np.ndarray:
    """
    Rotate points about the x-axis by the dip angle and then about the z-axis by the rotation angle.
    :param points: an array of points to rotate
    :param dip: the dip angle in radians
    :param rotation: the rotation angle in radians
    :return: the rotated points
    """
    # Assert points is a numpy array containing 3D points
    if not isinstance(points, np.ndarray) and points.ndim != 2 and points.shape[1] != 3:
        raise TypeError("Input points must be a 2D numpy array of shape (N, 3).")

    # rotate the points about the z-axis by the inverse rotation angle
    points = xy_rotation_matrix(-rotation) @ points.T

    # Rotate points with the dip angle
    points = yz_rotation_matrix(dip) @ points

    # Rotate back the points to initial orientation
    points = xy_rotation_matrix(rotation) @ points

    return points.T


def set_attributes(entity, **kwargs):
    """
    Loop over kwargs and set attributes to an entity.

    TODO: Deprecate in favor of explicit attribute setting.
    """
    for key, value in kwargs.items():
        try:
            setattr(entity, key, value)
        except AttributeError:
            continue


def map_name_attributes(object_, **kwargs: dict) -> dict:
    """
    Map attributes to an object. The object must have an '_attribute_map'.

    :param object_: The object to map the attributes to.
    :param kwargs: Dictionary of attributes.
    """
    mapping = getattr(object_, "_attribute_map", None)

    if mapping is None:
        raise AttributeError("Object must have an '_attribute_map' attribute.")

    new_args = {}
    for attr, item in kwargs.items():
        if attr in mapping:
            new_args[mapping[attr]] = item
        else:
            new_args[attr] = item

    return new_args


def map_attributes(object_, **kwargs):
    """
    Map attributes to an object. The object must have an '_attribute_map'.

    :param entity: The object to map the attributes to.
    :param kwargs: The kwargs to map to the object.
    """

    values = map_name_attributes(object_, **kwargs)  # Swap duplicates
    set_attributes(object_, **values)


def stringify(values: dict[str, Any]) -> dict[str, Any]:
    """
    Convert all values in a dictionary to string.

    :param values: Dictionary of values to be converted.
    """
    string_dict = {}
    for key, value in values.items():
        mappers = [nan2str, inf2str, as_str_if_uuid, none2str, path2str]
        string_dict[key] = dict_mapper(value, mappers)

    return string_dict


def to_list(value: Any) -> list:
    """
    Convert value to a list.

    :param value: The value to convert.

    :return: A list
    """
    # ensure the names are a list
    if isinstance(value, list):
        return value
    if isinstance(value, tuple):
        return list(value)
    return [value]


def to_tuple(value: Any) -> tuple:
    """
    Convert value to a tuple.

    :param value: The value to convert.

    :return: A tuple
    """
    # ensure the names are a tuple
    if isinstance(value, tuple):
        return value
    if isinstance(value, list):
        return tuple(value)
    return (value,)


class SetDict(dict):
    def __init__(self, **kwargs):
        kwargs = {k: self.make_set(v) for k, v in kwargs.items()}
        super().__init__(kwargs)

    def make_set(self, value):
        if isinstance(value, (set, tuple, list)):
            value = set(value)
        else:
            value = {value}
        return value

    def __setitem__(self, key, value):
        value = self.make_set(value)
        super().__setitem__(key, value)

    def update(self, value: dict, **kwargs) -> None:  # type: ignore
        for key, val in value.items():
            val = self.make_set(val)
            if key in self:
                val = self[key].union(val)
            value[key] = val
        super().update(value, **kwargs)


def inf2str(value):  # map np.inf to "inf"
    if not isinstance(value, (int, float)):
        return value
    return str(value) if not np.isfinite(value) else value


def list2str(value):
    if isinstance(value, list):  # & (key not in exclude):
        return str(value)[1:-1]
    return value


def none2str(value):
    if value is None:
        return ""
    return value


def path2str(value):
    if isinstance(value, Path):
        return str(value)
    return value


def nan2str(value):
    if value is np.nan:
        return ""
    return value


def str2none(value):
    if value == "":
        return None
    return value


def find_unique_name(name: str, names: list[str]) -> str:
    """
    Get a unique name not in a list of names.

    :param name: The name to check.
    :param names: The list of names to avoid.

    :return: a unique name.
    """

    if name not in names:
        return name

    count = 1
    while f"{name}({count})" in names:
        count += 1

    return f"{name}({count})"


def remove_duplicates_in_list(input_list: list) -> list:
    """
    Remove duplicates from a list without changing the sorting.

    :param input_list: the list to remove duplicates from.

    :return: The sorted list
    """
    return sorted(set(input_list), key=input_list.index)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from enum import Enum
from pathlib import Path

import numpy as np


class PrimitiveTypeEnum(Enum):
    """
    Known data type.

    Available options:
    """

    INVALID = 0
    INTEGER = 1
    FLOAT = 2
    TEXT = 3
    REFERENCED = 4
    FILENAME = 5
    BLOB = 6
    VECTOR = 7
    DATETIME = 8
    GEOMETRIC = 9
    MULTI_TEXT = 10
    BOOLEAN = 11


class DataTypeEnum(Enum):
    INVALID = type(None)
    INTEGER = np.int32
    FLOAT = np.float32
    TEXT = str
    REFERENCED = np.uint32  # Could represent a reference type
    FILENAME = Path
    BLOB = bytes
    VECTOR = type(None)  # Assuming a vector is a list
    DATETIME = str  # Could use datetime
    GEOMETRIC = type(None)  # For custom geometric type
    MULTI_TEXT = str
    BOOLEAN = bool

    @classmethod
    def from_primitive_type(cls, primitive_type: PrimitiveTypeEnum) -> type:
        """
        Get the data type from the primitive type.

        :param primitive_type: The primitive type.
        :return: The data type.
        """
        return DataTypeEnum[primitive_type.name].value



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import json
from pathlib import Path
from typing import Any
from uuid import UUID

from pydantic import BaseModel, ConfigDict, field_validator

from geoh5py import Workspace
from geoh5py.shared.utils import fetch_active_workspace
from geoh5py.ui_json.forms import BaseForm
from geoh5py.ui_json.validation import ErrorLane, ErrorPool, UIJsonError


class BaseUIJson(BaseModel):
    """
    Base class for storing ui.json data on disk.

    :params title: Title of the application.
    :params geoh5: Path to the geoh5 file.
    :params run_command: Command to run the application.
    :params run_command_boolean: Boolean to run the command.
    :params monitoring_directory: Directory to monitor for changes.
    :params conda_environment: Conda environment to run the application.
    :params conda_environment_boolean: Boolean to run the conda environment.
    :params workspace_geoh5: Path to the workspace geoh5 file.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    title: str
    geoh5: Path | None
    run_command: str
    monitoring_directory: Path
    conda_environment: str
    workspace_geoh5: Path

    @field_validator("workspace_geoh5", mode="after")
    @classmethod
    def current_directory_if_workspace_doesnt_exist(cls, path):
        if not path.exists():
            return Path()
        return path

    @field_validator("geoh5", mode="after")
    @classmethod
    def workspace_path_exists(cls, path: Path):
        if not path.exists():
            raise FileNotFoundError(f"geoh5 path {path} does not exist.")
        return path

    @classmethod
    def read(cls, path: Path):
        """Create a UIJson object from file."""

        path = path.resolve()

        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist.")

        if "".join(path.suffixes[-2:]) != ".ui.json":
            raise ValueError(f"File {path} is not a .ui.json file.")

        with open(path, encoding="utf-8") as file:
            kwargs = json.load(file)

        return cls(**kwargs)

    def to_params(self, workspace: Workspace | None = None) -> dict[str, Any]:
        """
        Promote, flatten and validate parameter/values dictionary.

        :param workspace: Workspace to fetch entities from.  Used for passing active
            workspaces to avoid closing and flushing data.
        """

        with fetch_active_workspace(workspace or Workspace(self.geoh5)) as geoh5:
            if geoh5 is None:
                raise ValueError("Workspace cannot be None.")

            data = {}
            for field in self.model_fields_set:
                if field == "geoh5":
                    data[field] = geoh5
                    continue

                value = getattr(self, field)
                value = value.flatten() if isinstance(value, BaseForm) else value

                if isinstance(value, UUID):
                    value = self._object_or_catch(geoh5, value)

                data[field] = value

            self.validate_data(data)

        return data

    def validate_data(self, params: dict[str, Any] | None = None):
        """
        Validate the UIJson data.

        :param params: promoted and flattened parameters/values dictionary.  The params
            dictionary will be generated from the model values if not provided.

        :raises UIJsonError: If any validations fail.
        """

        if params is None:
            self.to_params()
            return

        for field in self.model_fields_set:
            value = getattr(self, field)
            if not isinstance(value, BaseForm):
                continue

            try:
                value.validate_data(params)
            except UIJsonError as e:
                if isinstance(params[field], ErrorLane):
                    params[field].catch(e)
                else:
                    params[field] = ErrorLane(e)

        error_pool = ErrorPool(
            {k: v for k, v in params.items() if isinstance(v, ErrorLane)}
        )
        error_pool.throw()

    def _object_or_catch(
        self,
        workspace: Workspace,
        uuid: UUID,
    ):
        """
        Returns an object if it exists in the workspace or an error if not.

        :param workspace: Workspace to fetch entities from.
        :param uuid: UUID of the object to fetch.
        """

        obj = workspace.get_entity(uuid)
        if obj[0] is not None:
            return obj[0]

        error_lane = ErrorLane(
            UIJsonError(f"Workspace does not contain an entity with uid: {uuid}.")
        )

        return error_lane



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import Any

from .data import Data, PrimitiveTypeEnum


class BlobData(Data):
    @classmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        return PrimitiveTypeEnum.BLOB

    def validate_values(self, values: Any | None) -> Any:
        """
        Validate values for BlobData.
        """
        if values is not None:
            if not isinstance(values, bytes):
                raise TypeError(f"Values must be of type bytes. Got {type(values)}")
        return values

    # TODO: implement specialization to access values.
    # Stored as a 1D array of 8-bit char type (native) (value '0' or '1').
    # For each index set to 1, an opaque data set named after the index (e.g. "1", "2", etc)
    # must be added under the Data instance, containing the binary data tied to that index.
    # No data value : 0



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
from numbers import Real

import numpy as np

from ..objects import GeoImage
from ..shared.conversion import Grid2DConversion
from ..shared.utils import mask_by_extent, xy_rotation_matrix, yz_rotation_matrix
from .grid_object import GridObject


class Grid2D(GridObject):
    """
    Rectilinear 2D grid of uniform cell size. The grid can
    be oriented in 3D space through horizontal :obj:`~geoh5py.objects.grid2d.Grid2D.rotation`
    and :obj:`~geoh5py.objects.grid2d.Grid2D.dip` parameters.
    Nodal coordinates are determined relative to the origin and the sign
    of cell delimiters.

    :param u_cell_size: Cell size along the u-axis.
    :param v_cell_size: Cell size along the v-axis.
    :param u_count: Number of cells along the u-axis.
    :param v_count: Number of cells along the v-axis.
    :param vertical: Set the grid to be vertical.
    :param dip: Dip angle from horizontal (positive down) in degrees.
        Defaults to an horizontal grid (dip=0).
    """

    _TYPE_UID = uuid.UUID(
        fields=(0x48F5054A, 0x1C5C, 0x4CA4, 0x90, 0x48, 0x80F36DC60A06)
    )
    _attribute_map = GridObject._attribute_map.copy()
    _attribute_map.update(
        {
            "Dip": "dip",
            "U Count": "u_count",
            "V Count": "v_count",
            "Origin": "origin",
            "Rotation": "rotation",
            "U Size": "u_cell_size",
            "V Size": "v_cell_size",
            "Vertical": "vertical",
        }
    )
    _converter: type[Grid2DConversion] = Grid2DConversion

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        u_cell_size: float = 1.0,
        v_cell_size: float = 1.0,
        u_count: int = 1,
        v_count: int = 1,
        vertical: bool = False,
        dip: float = 0.0,
        **kwargs,
    ):
        self._u_count: np.int32 = self.validate_count(u_count, "u")
        self._v_count: np.int32 = self.validate_count(v_count, "v")

        super().__init__(
            **kwargs,
        )

        self.u_cell_size = u_cell_size
        self.v_cell_size = v_cell_size
        self.dip = dip
        self.vertical = vertical

    @property
    def cell_center_u(self) -> np.ndarray:
        """
        Cell center local coordinate along the u-axis,
        shape(:obj:`~geoh5py.objects.grid2d.Grid2D.u_count`, )
        """
        return (
            np.cumsum(np.ones(self.u_count) * self.u_cell_size) - self.u_cell_size / 2.0
        )

    @property
    def cell_center_v(self) -> np.ndarray:
        """
        The cell center local coordinate along the v-axis,
        shape(:obj:`~geoh5py.objects.grid2d.Grid2D.u_count`, )
        """
        return (
            np.cumsum(np.ones(self.v_count) * self.v_cell_size) - self.v_cell_size / 2.0
        )

    @property
    def centroids(self) -> np.ndarray:
        """
        Cell center locations in world coordinates,
        shape(:obj:`~geoh5py.objects.grid2d.Grid2D.n_cells`, 3).

        .. code-block:: python

            centroids = [
                [x_1, y_1, z_1],
                ...,
                [x_N, y_N, z_N]
            ]
        """
        if getattr(self, "_centroids", None) is None:
            rotation_matrix = xy_rotation_matrix(np.deg2rad(self.rotation))
            dip_matrix = yz_rotation_matrix(np.deg2rad(self.dip))

            u_grid, v_grid = np.meshgrid(self.cell_center_u, self.cell_center_v)
            xyz = np.c_[np.ravel(u_grid), np.ravel(v_grid), np.zeros(self.n_cells)]

            xyz_dipped = dip_matrix @ xyz.T
            centroids = (rotation_matrix @ xyz_dipped).T

            for ind, axis in enumerate(["x", "y", "z"]):
                centroids[:, ind] += self.origin[axis]

            self._centroids = centroids

        return self._centroids

    def copy_from_extent(  # pylint: disable=too-many-locals
        self,
        extent: np.ndarray,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        inverse: bool = False,
        **kwargs,
    ) -> Grid2D | None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.copy_from_extent`.
        """
        if not isinstance(extent, np.ndarray):
            raise TypeError("Expected a numpy array of extent values.")

        if not extent.ndim == 2 or extent.shape not in [(2, 3), (2, 2)]:
            raise TypeError("Expected a 2D numpy array with 2 or 3 columns")

        # get the centroids
        selected_centroids = mask_by_extent(
            self.centroids, extent, inverse=inverse
        ).reshape((self.v_count, self.u_count))

        u_ind = np.any(selected_centroids, axis=0)
        v_ind = np.any(selected_centroids, axis=1)

        indices = np.kron(v_ind, u_ind).flatten()

        if not np.any(indices):
            return None

        if not inverse:
            delta_orig = np.c_[
                np.argmax(u_ind) * self.u_cell_size,
                np.argmax(v_ind) * self.v_cell_size,
                0.0,
            ].T

            dip_matrix = yz_rotation_matrix(np.deg2rad(self.dip))
            delta_orig = dip_matrix @ delta_orig

            rotation_matrix = xy_rotation_matrix(np.deg2rad(self.rotation))
            delta_orig = (rotation_matrix @ delta_orig).T

            kwargs.update(
                {
                    "origin": np.r_[
                        self.origin["x"] + delta_orig[0, 0],
                        self.origin["y"] + delta_orig[0, 1],
                        self.origin["z"] + delta_orig[0, 2],
                    ],
                    "u_count": np.sum(u_ind),
                    "v_count": np.sum(v_ind),
                }
            )

        else:
            indices = GridObject.mask_by_extent(self, extent, inverse=inverse)

        copy = super(GridObject, self).copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            mask=indices,
            **kwargs,
        )

        if not inverse:
            for child in copy.children:
                if isinstance(getattr(child, "values", None), np.ndarray):
                    indices = child.mask_by_extent(extent, inverse=inverse)
                    values = child.values
                    values[~indices] = child.nan_value
                    child.values = values

        return copy

    @property
    def dip(self) -> float:
        """
        Dip angle from horizontal (positive down) in degrees.
        """
        if self.vertical:
            self._dip = 90.0
            return self._dip

        return self._dip

    @dip.setter
    def dip(self, value: Real):
        if not isinstance(value, Real):
            raise TypeError("Dip angle must be a float.")

        self._centroids = None

        self._dip = float(value)
        if self._dip == 90:
            self._vertical = True

        self.workspace.update_attribute(self, "attributes")

    @property
    def shape(self) -> tuple[np.int32, np.int32]:
        """
        Number of cells along the u and v-axis.
        """
        return self.u_count, self.v_count

    @property
    def u_cell_size(self) -> float:
        """
        Cell size along the u-axis.
        """
        return self._u_cell_size

    @u_cell_size.setter
    def u_cell_size(self, value: Real | np.ndarray):
        self._u_cell_size = self.validate_cell_size(value, "u")
        self._centroids = None

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def u_count(self) -> np.int32:
        """
        Number of cells along v-axis.
        """
        return self._u_count

    @property
    def v_cell_size(self) -> float:
        """
        Cell size along the v-axis
        """
        return self._v_cell_size

    @v_cell_size.setter
    def v_cell_size(self, value: Real | np.ndarray):
        self._v_cell_size = self.validate_cell_size(value, "v")
        self._centroids = None

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def v_count(self) -> np.int32:
        """
        Number of cells along v-axis.
        """
        return self._v_count

    @staticmethod
    def validate_cell_size(value: Real | np.ndarray, axis: str) -> float:
        """
        Validate and format type of cell size value.
        """
        if not isinstance(value, (Real, np.ndarray)):
            raise TypeError(f"Attribute '{axis}_cell_size' must be type(float).")

        if isinstance(value, np.ndarray):
            if not len(value) == 1:
                raise ValueError(
                    "Attribute 'v_cell_size' must be a float of shape (1,)"
                )

            return np.r_[value].astype(float).item()

        return float(value)

    @property
    def vertical(self) -> bool:
        """
        Set the grid to be vertical.
        """
        return self._vertical

    @vertical.setter
    def vertical(self, value: bool):
        if not isinstance(value, bool) is not None or value not in [
            0,
            1,
        ]:
            raise TypeError("Attribute 'vertical' must be type(bool).")

        self._centroids = None
        self._vertical = bool(value)

        if self.dip != 90 and value is True:
            self._dip = 90.0

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    def to_geoimage(
        self, keys: list | str, mode: str | None = None, **geoimage_kwargs
    ) -> GeoImage:
        """
        Create a :obj:geoh5py.objects.geo_image.GeoImage object from the current Grid2D.

        :param keys: the list of the data name to pass as band in the image.
            Warning: The len of the list can only be 1, 3, 4 (Pillow restrictions).
        :param mode: The mode of the image. One of 'GRAY', 'RGB', 'RGBA' or 'CMYK'.

        :return: a new georeferenced :obj:`geoh5py.objects.geo_image.GeoImage`.
        """
        return self.converter.to_geoimage(self, keys, mode=mode, **geoimage_kwargs)

    @staticmethod
    def validate_count(value: int, axis: str) -> np.int32:
        """
        Validate and format type of count value.
        """
        if not isinstance(value, (np.integer, int)) or value < 1:
            raise TypeError(
                f"Attribute '{axis}_count' must be a type(int32) greater than 1."
            )

        return np.int32(value)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import numpy as np

from ...objects import CellObject, Curve, Surface
from ...workspace import Workspace
from .points import PointsMerger


class CellMerger(PointsMerger):
    _type: type = CellObject

    @classmethod
    def create_object(
        cls, workspace: Workspace, input_entities: list, **kwargs
    ) -> CellObject:
        """
        Create a new object of type cls._type from a list of input entities.
        It merges the cells together and create a new object with the merged cells.

        :param workspace: The workspace to create the object in.
        :param input_entities: The list of input entities to merge together.
        :param kwargs: The kwargs to pass to the object creation.

        :return: The newly created object merged from input_entities.
        """
        # create the vertices
        vertices = np.vstack([input_entity.vertices for input_entity in input_entities])

        # merge the simplices
        cells: list = []
        previous: int = 0
        for entity in input_entities:
            temp_cells = entity.cells + previous
            cells.append(temp_cells)
            previous = np.nanmax(temp_cells) + 1

        # create an object of type
        output = cls._type.create(  # type: ignore
            workspace, vertices=vertices, cells=np.vstack(cells).tolist(), **kwargs
        )

        return output


class CurveMerger(CellMerger):
    _type = Curve


class SurfaceMerger(CellMerger):
    _type = Surface



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING
from uuid import UUID

import numpy as np

from ..data import DataAssociationEnum
from ..data.primitive_type_enum import DataTypeEnum


if TYPE_CHECKING:  # pragma: no cover
    from .property_group import PropertyGroup


class PropertyGroupTable(ABC):
    """
    A class to store the information of a property group.

    :param property_group: The property group to extract the data from.
    """

    locations_keys = ("X", "Y", "Z")

    def __init__(self, property_group: PropertyGroup):
        if not hasattr(property_group, "property_group_type"):
            raise TypeError("'property_group' must be a PropertyGroup object.")

        if hasattr(property_group.parent, "collar"):
            raise NotImplementedError(
                "PropertyGroupTable is not supported for Drillhole objects."
            )

        self._property_group: PropertyGroup = property_group

    def __call__(
        self, spatial_index: bool = False, use_uids: bool = False
    ) -> np.ndarray | None:
        """
        Create a structured array with the data of the properties.

        :param spatial_index: If True, the spatial index is added to the table.
        :param use_uids: If True, the uids are used as columns name.

        :return: A table with the data of the properties.
        """
        if (
            self.property_group.properties is None
            or self.property_group.properties_name is None
        ):
            return None

        keys: list[UUID] = self.property_group.properties
        names: list[str] | list[UUID] = (
            self.property_group.properties
            if use_uids
            else self.property_group.properties_name
        )

        output_array = self._create_empty_structured_array(names, keys, spatial_index)

        if spatial_index:
            for idx, key in enumerate(self.locations_keys):
                output_array[key] = self.locations[:, idx]

        for key, name in zip(keys, names, strict=False):  # type: ignore
            data = self.property_group.parent.get_data(key)[0]
            output_array[str(name)] = data.values

        return output_array

    def _create_empty_structured_array(
        self,
        properties_name: list[str] | list[UUID],
        properties_keys: list[UUID],
        spatial_index: bool = False,
    ) -> np.ndarray:
        """
        Create an empty structured array that can contains the data.

        :param properties_name: The names of the properties.
        :param properties_keys: The keys of the properties.
        :param spatial_index: If True, the association is added to the table.

        :return: an empty structured array.
        """
        dtypes = (
            [(loc, np.float32) for loc in self.locations_keys] if spatial_index else []
        )

        for key, name in zip(properties_keys, properties_name, strict=False):
            data = self.property_group.parent.get_data(key)[0]
            dtype: type | str = DataTypeEnum.from_primitive_type(
                data.entity_type.primitive_type
            )
            if dtype not in [np.float32, np.int32, np.uint32, bool]:
                dtype = "O"
            dtypes.append((str(name), dtype))

        empty_array = np.recarray((self.size,), dtype=dtypes)

        return empty_array

    @property
    def locations(self) -> np.ndarray:
        """
        The locations of the association table.

        This function is needed as a data can be both associated to cell or
        vertex in CellObjects.
        """
        if self.property_group.association == DataAssociationEnum.VERTEX:
            return self.property_group.parent.vertices  # type: ignore

        if self.property_group.association == DataAssociationEnum.CELL:
            return self.property_group.parent.centroids  # type: ignore

        raise ValueError(
            f"The association {self.property_group.association} is not supported. "
            f"Only VERTEX and CELL associations are supported."
        )

    @property
    def property_group(self) -> PropertyGroup:
        """
        The property group to extract the data from.
        """
        return self._property_group

    @property
    def size(self) -> int:
        """
        The size of the properties in the group.
        """
        return self.locations.shape[0]



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any
from uuid import UUID

from pydantic import BaseModel, Field

from ..shared.utils import str2uuid, to_tuple
from .base import Group


if TYPE_CHECKING:  # pragma: no cover
    from ..objects import Curve, ObjectBase, Slicer


class InterpretationSectionParams(BaseModel):
    """
    The class to store the interpretation section.

    :param normal_x: The normal x component.
    :param normal_y: The normal y component.
    :param normal_z: The normal z component.
    :param position_x: The position x component.
    :param position_y: The position y component.
    :param position_z: The position z component.
    """

    normal_x: float = Field(alias="Normal X")
    normal_y: float = Field(alias="Normal Y")
    normal_z: float = Field(alias="Normal Z")
    position_x: float = Field(alias="Position X")
    position_y: float = Field(alias="Position Y")
    position_z: float = Field(alias="Position Z")

    @classmethod
    def create_section(
        cls,
        params: dict | InterpretationSectionParams,
    ) -> InterpretationSectionParams:
        """
        Verify/Create the interpretation section from a mapper.

        :param params: The parameters to verify.

        :return: An instance of the class.
        """
        if isinstance(params, Mapping):
            return cls(**params)
        if isinstance(params, InterpretationSectionParams):
            return params
        raise TypeError(
            "Interpretation section must be an InterpretationSectionParams"
            f"or a dict, {type(params)} provided."
        )


class InterpretationSection(Group):
    """The type for the basic Container group."""

    _attribute_map = Group._attribute_map.copy()
    _attribute_map.update(
        {
            "Can add group": "can_add_group",
        }
    )

    _TYPE_UID = UUID(fields=(0x27EE59E1, 0xB1CE, 0x41EC, 0x8A, 0x86, 0x8BD3D229E198))
    _default_name = "Interpretation Section"

    def __init__(self, can_add_group: bool = False, **kwargs):
        self._can_add_group: bool = self._validate_can_add_group(can_add_group)

        super().__init__(**kwargs)

        self._interpretation_curves: tuple[Curve] | tuple = ()
        self._interpretation_sections: tuple[InterpretationSectionParams] | tuple = ()
        self._section_object: Slicer | None = None

    def _load_from_metadata(self, key: str) -> Any:
        """
        Load the metadata from the key.
        :param key: the key to load the metadata from.
        :return: The value store in the metadata if exists.
        """
        return self.metadata.get(key, None) if self.metadata else None

    def _update_to_metadata(self, key: str, values: tuple):
        """
        Update the metadata with the interpretation curves.

        :param key: The key to update.
        :param values: The values to update.
        """
        metadata: None | list = None

        if values:
            if key == "Interpretation curves":
                metadata = [value.uid for value in values]
            elif key == "Interpretation sections":
                metadata = [value.model_dump(by_alias=True) for value in values]
            else:
                raise ValueError(f"Invalid key '{key}' provided.")

        self.update_metadata({key: metadata})

    @staticmethod
    def _validate_can_add_group(value: bool) -> bool:
        """
        Validate the can add group attribute.

        :param value: A boolean value.

        :return: the verified and converted value.
        """
        if value not in [True, False, 0, 1]:
            raise TypeError("'Can add group' must be a boolean value.")

        return bool(value)

    def _verify_object(self, value, key: str) -> ObjectBase:
        """
        Get the object and verify that it is a children of the group.

        :param value: The value to verify.
        :param key: The default name of the expected object type.

        :return: The verified object
        """
        if isinstance(value, str | UUID):
            value = self.workspace.get_entity(str2uuid(value))[0]
        if getattr(value, "_default_name", None) != key:
            raise TypeError(
                f"The {value} object must be a {key} object."
                f" {type(value)} provided."
            )

        return value

    def add_interpretation_curve(self, curves: Curve | tuple[Curve]):
        """
        Add an interpretation curve.

        :param curves: The interpretation curve to add.
        """
        if curves is None:
            return

        for curve in to_tuple(curves):
            curve = self._verify_object(curve, "Curve")
            self.add_children(curve)
            self._interpretation_curves += (curve,)

            if self.section_object is not None:
                curve.clipping_ids = [self.section_object.uid]

        self._update_to_metadata("Interpretation curves", self._interpretation_curves)

    def add_interpretation_section(
        self, sections: InterpretationSectionParams | list[InterpretationSectionParams]
    ):
        """
        Add an interpretation section.

        :param sections: The interpretation section to add.
        """
        if sections is None:
            return

        for section in to_tuple(sections):
            self._interpretation_sections += (
                InterpretationSectionParams.create_section(section),
            )

        self._update_to_metadata(
            "Interpretation sections", self._interpretation_sections
        )

    @property
    def can_add_group(self) -> bool:
        """
        Get the can add group attribute.

        :return: The 'can add group' attribute.
        """
        return self._can_add_group

    @can_add_group.setter
    def can_add_group(self, value: bool):
        """
        Set the can add group attribute.

        :param value: The can add group attribute.
        """
        self._can_add_group = self._validate_can_add_group(value)
        self.workspace.update_attribute(self, "attributes")

    @property
    def interpretation_curves(self) -> tuple[Curve]:
        """
        Get the interpretation curves.

        :return: The interpretation curves.
        """
        if not self._interpretation_curves:
            self.add_interpretation_curve(
                self._load_from_metadata("Interpretation curves")
            )
        return self._interpretation_curves

    @property
    def interpretation_sections(self) -> tuple[InterpretationSectionParams] | None:
        """
        Get the interpretation sections.

        :return: The interpretation sections.
        """
        if not self._interpretation_sections:
            self.add_interpretation_section(
                self._load_from_metadata("Interpretation sections")
            )

        return self._interpretation_sections

    def remove_interpretation_curve(self, curves: Curve | tuple[Curve]):
        """
        Remove an interpretation curve.

        :param curves: The interpretation curve to remove.
        """
        if not self.interpretation_curves:
            return

        verified = [self._verify_object(curve, "Curve") for curve in to_tuple(curves)]

        interpretation_curves = []
        for curve in self.interpretation_curves:
            if curve in verified:
                self.workspace.remove_entity(curve)
            else:
                interpretation_curves.append(curve)

        self._interpretation_curves = tuple(interpretation_curves)
        self._update_to_metadata("Interpretation curves", self._interpretation_curves)

    def remove_interpretation_section(
        self, sections: InterpretationSectionParams | tuple[InterpretationSectionParams]
    ):
        """
        Remove an interpretation section.

        :param sections: The interpretation section to remove.
        """
        if not self.interpretation_sections:
            return

        verified = [
            InterpretationSectionParams.create_section(section)
            for section in to_tuple(sections)
        ]
        self._interpretation_sections = tuple(
            section
            for section in self.interpretation_sections
            if section not in verified
        )
        self._update_to_metadata(
            "Interpretation sections", self._interpretation_sections
        )

    @property
    def section_object(self) -> Slicer | None:
        """
        Get the section object ID.

        :return: The section object ID.
        """
        if self._section_object is None:
            self.section_object = self._load_from_metadata("Section object ID")

        return self._section_object

    @section_object.setter
    def section_object(self, slicer: Slicer | None):
        """
        Set the section object ID.

        :param slicer: The section object ID.
        """
        original_slicer = self._section_object
        if slicer is None:
            self._section_object = None
            self.update_metadata({"Section object ID": None})
        else:
            # a lot of type ignore because of circular imports
            self._section_object = self._verify_object(slicer, "Slicer")  # type: ignore
            self.add_children(self._section_object)  # type: ignore
            self.clipping_ids = [self._section_object.uid]  # type: ignore
            self.update_metadata({"Section object ID": self._section_object.uid})  # type: ignore

        if original_slicer is not None:
            self.workspace.remove_entity(original_slicer)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
from abc import ABC

from geoh5py.objects.curve import Curve

from .base import AirborneEMSurvey, FEMSurvey


# pylint: disable=too-many-ancestors


class AirborneFEMSurvey(FEMSurvey, AirborneEMSurvey, ABC):
    @property
    def base_receiver_type(self):
        """
        :return: Base receiver class
        """
        return Curve

    @property
    def base_transmitter_type(self):
        """
        :return: Base transmitter class
        """
        return Curve

    @property
    def default_metadata(self) -> dict:
        """
        Default dictionary of metadata for AirborneFEM entities.
        """
        return {
            "EM Dataset": {
                "Channels": [],
                "Input type": "Rx",
                "Property groups": [],
                "Receivers": None,
                "Survey type": "Airborne FEM",
                "Transmitters": None,
                "Unit": "Hertz (Hz)",
            }
        }

    @property
    def default_receiver_type(self):
        """
        :return: Transmitter class
        """
        return AirborneFEMReceivers

    @property
    def default_transmitter_type(self):
        """
        :return: Transmitter class
        """
        return AirborneFEMTransmitters


class AirborneFEMReceivers(AirborneFEMSurvey):  # pylint: disable=too-many-ancestors
    """
    Airborne frequency-domain electromagnetic receivers class.
    """

    _TYPE_UID = uuid.UUID("{b3a47539-0301-4b27-922e-1dde9d882c60}")
    __TYPE = "Receivers"
    _default_name = "Airborne FEM Rx"

    @property
    def complement(self):
        return self.transmitters

    @property
    def type(self):
        """Survey element type"""
        return self.__TYPE


class AirborneFEMTransmitters(AirborneFEMSurvey):  # pylint: disable=too-many-ancestors
    """
    Airborne frequency-domain electromagnetic transmitters class.
    """

    _TYPE_UID = uuid.UUID("{a006cf3e-e24a-4c02-b904-2e57b9b5916d}")
    __TYPE = "Transmitters"
    _default_name = "Airborne FEM Tx"

    @property
    def complement(self):
        return self.receivers

    @property
    def type(self):
        """Survey element type"""
        return self.__TYPE



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .concatenator import Concatenator


class Concatenated:
    """
    Base class modifier for concatenated objects and data.
    """

    _concat_attr_str: str = "Attributes"

    def __init__(self, *args, **kwargs):
        attribute_map = getattr(self, "_attribute_map", {})
        attr = {"name": "Entity", "parent": None}

        for key, value in kwargs.items():
            attr[attribute_map.get(key, key)] = value

        super().__init__(*args, **attr)

    @property
    def concat_attr_str(self) -> str:
        """String identifier for the concatenated attributes."""
        return self._concat_attr_str

    @property
    def concatenator(self) -> Concatenator:
        """
        Parental Concatenator entity.

        Warning: Parent is not an attribute of Concatenated, but of the derived class.
        """
        parent: Concatenated | Concatenator = getattr(self, "parent", None)  # type: ignore
        if parent is None:
            raise UserWarning("Parent of concatenated object is not defined.")

        if isinstance(parent, Concatenated):
            return parent.concatenator

        return parent



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
from abc import abstractmethod
from typing import Any

import numpy as np

from ..shared import Entity
from ..shared.utils import mask_by_extent
from .data_association_enum import DataAssociationEnum
from .data_type import DataType, ReferenceDataType
from .primitive_type_enum import PrimitiveTypeEnum


class Data(Entity):
    """
    Base class for Data entities.

    :param association: Relationship made between the parent object and values.
    :param modifiable: Entity can be modified.
    :param visible: Entity is visible. Only a single data can be visible at a time.
    :param values: Data values.
    """

    _attribute_map = Entity._attribute_map.copy()
    _attribute_map.update({"Association": "association", "Modifiable": "modifiable"})

    def __init__(
        self,
        association: DataAssociationEnum = DataAssociationEnum.OBJECT,
        modifiable: bool = True,
        visible: bool = False,
        values: Any | None = None,
        **kwargs,
    ):
        self._association = self.validate_association(association)

        super().__init__(visible=visible, **kwargs)

        self.modifiable = modifiable
        self.values = values

    def copy(
        self,
        parent=None,
        *,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ) -> Data:
        """
        Function to copy data to a different parent entity.

        :param parent: Target parent to copy the entity under. Copied to current
            :obj:`~geoh5py.shared.entity.Entity.parent` if None.
        :param clear_cache: Clear array attributes after copy.
        :param mask: Array of bool defining the values to keep.
        :param kwargs: Additional keyword arguments to pass to the copy constructor.

        :return entity: Registered Entity to the workspace.
        """
        if parent is None:
            parent = self.parent

        if (
            self.values is not None
            and mask is not None
            and self.association
            in (
                DataAssociationEnum.VERTEX,
                DataAssociationEnum.CELL,
            )
        ):
            if not isinstance(mask, np.ndarray):
                raise TypeError("Mask must be an array or None.")

            if mask.dtype != bool or mask.shape != self.values.shape:
                raise ValueError(
                    f"Mask must be a boolean array of shape {self.values.shape}, not {mask.shape}"
                )

            n_values = (
                parent.n_cells
                if self.association is DataAssociationEnum.CELL
                else parent.n_vertices
            )

            # Trimming values to the mask
            if n_values < self.values.shape[0]:
                values = self.values[mask]
            else:
                values = self.values.copy()
                values[~mask] = self.nan_value

            kwargs.update({"values": values})

        new_entity = parent.workspace.copy_to_parent(
            self,
            parent,
            clear_cache=clear_cache,
            **kwargs,
        )

        return new_entity

    def copy_from_extent(
        self,
        extent: np.ndarray,
        parent=None,
        *,
        clear_cache: bool = False,
        inverse: bool = False,
        **kwargs,
    ) -> Entity | None:
        """
        Function to copy data based on a bounding box extent.

        :param extent: Bounding box extent requested for the input entity, as supplied for
            :func:`~geoh5py.shared.entity.Entity.mask_by_extent`.
        :param parent: Target parent to copy the entity under. Copied to current
            :obj:`~geoh5py.shared.entity.Entity.parent` if None.
        :param clear_cache: Clear array attributes after copy.
        :param inverse: Keep the inverse (clip) of the extent selection.
        :param kwargs: Additional keyword arguments to pass to the copy constructor.

        :return entity: Registered Entity to the workspace.
        """
        indices = self.mask_by_extent(extent, inverse=inverse)
        if indices is None:
            return None

        return self.copy(
            parent=parent,
            clear_cache=clear_cache,
            mask=indices,
            **kwargs,
        )

    @classmethod
    def default_type_uid(cls) -> uuid.UUID | None:
        """Abstract method to return the default type uid for the class."""
        return None

    @property
    def formatted_values(self):
        """
        The data with the right format for Geoscience ANALYST.
        """
        return self.values

    @property
    def n_values(self) -> int | None:
        """
        Number of expected data values based on :obj:`~geoh5py.data.data.Data.association`
        """
        if self.association in [
            DataAssociationEnum.VERTEX,
            DataAssociationEnum.DEPTH,
        ] and hasattr(self.parent, "n_vertices"):
            return self.parent.n_vertices
        if self.association is DataAssociationEnum.CELL and hasattr(
            self.parent, "n_cells"
        ):
            return self.parent.n_cells
        if self.association is DataAssociationEnum.FACE and hasattr(
            self.parent, "n_faces"
        ):
            return self.parent.n_faces
        if self.association is DataAssociationEnum.OBJECT:
            return 1

        return None

    @property
    def nan_value(self) -> None:
        """
        Value used to represent missing data in python.
        """
        return None

    @property
    def association(self) -> DataAssociationEnum:
        """
        Relationship made between the
        :func:`~geoh5py.data.data.Data.values` and elements of the
        :obj:`~geoh5py.shared.entity.Entity.parent` object.
        """
        return self._association

    @property
    def entity_type(self):
        """
        Type of data.
        """
        return self._entity_type

    @entity_type.setter
    def entity_type(self, data_type: DataType | ReferenceDataType):
        self._entity_type = data_type

        if self.on_file:
            self.workspace.update_attribute(self, "entity_type")

    @property
    def modifiable(self) -> bool:
        """
        Entity can be modified within ANALYST.
        """
        return self._modifiable

    @modifiable.setter
    def modifiable(self, value: bool):
        self._modifiable = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @classmethod
    @abstractmethod
    def primitive_type(cls) -> PrimitiveTypeEnum:
        """Abstract method to return the primitive type of the class."""

    def mask_by_extent(
        self, extent: np.ndarray, inverse: bool = False
    ) -> np.ndarray | None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.mask_by_extent`.

        Uses the parent object's vertices or centroids coordinates.

        :param extent: Array or coordinate defining the lower and upper bounds of the extent.
        :param inverse: Keep the inverse (clip) of the extent selection.
        """
        if self.association is DataAssociationEnum.VERTEX and hasattr(
            self.parent, "vertices"
        ):
            return mask_by_extent(self.parent.vertices, extent, inverse=inverse)

        if self.association is DataAssociationEnum.CELL:
            if hasattr(self.parent, "vertices") and hasattr(self.parent, "_cells"):
                indices = mask_by_extent(self.parent.vertices, extent, inverse=inverse)
                if indices is not None:
                    indices = np.all(indices[self.parent.cells], axis=1)

                return indices

            if hasattr(self.parent, "centroids"):
                return mask_by_extent(self.parent.centroids, extent, inverse=inverse)

        return None

    def validate_entity_type(self, entity_type: DataType | None) -> DataType:
        """
        Validate the entity type.

        :param entity_type: Entity type to validate.
        """
        if (
            not isinstance(entity_type, DataType)
            or entity_type.primitive_type != self.primitive_type()
        ):
            raise TypeError(
                "Input 'entity_type' must be a DataType object of primitive_type 'TEXT'."
            )

        if entity_type.name == "Entity":
            entity_type.name = self.name

        return entity_type

    @staticmethod
    def validate_association(value: str | DataAssociationEnum):
        if isinstance(value, str):
            if value.upper() not in DataAssociationEnum.__members__:
                raise ValueError(
                    f"Association flag should be one of {DataAssociationEnum.__members__}"
                )

            value = getattr(DataAssociationEnum, value.upper())

        if not isinstance(value, DataAssociationEnum):
            raise TypeError(f"Association must be of type {DataAssociationEnum}")

        return value

    @abstractmethod
    def validate_values(self, values: Any | None) -> Any:
        """
        Validate the values.

        To be deprecated along with the standalone Drillhole class in future version.

        :param values: Values to validate.
        """

    @property
    def values(self):
        """
        Data values
        """
        if getattr(self, "_values", None) is None:
            values = self.workspace.fetch_values(self)
            self._values = self.validate_values(values)

        return self._values

    @values.setter
    def values(self, values: np.ndarray | None):
        self._values = self.validate_values(values)

        if self.on_file:
            self.workspace.update_attribute(self, "values")

    @property
    def visible(self) -> bool:
        """
        Whether the data is visible in camera (checked in ANALYST object tree).
        """
        return self._visible

    @visible.setter
    def visible(self, value: bool):
        self._visible = value

        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

            if value:
                for child in self.parent.children:
                    child.visible = False

    def __call__(self):
        return self.values



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid
import warnings

import numpy as np

from ..data import DataAssociationEnum
from .object_base import ObjectBase


class Points(ObjectBase):
    """
    Points object made up of vertices.

    :param vertices: Array of vertices coordinates, shape(n_vertices, 3).
    """

    _default_name = "Points"
    _TYPE_UID: uuid.UUID | None = uuid.UUID("{202C5DB1-A56D-4004-9CAD-BAAFD8899406}")
    __VERTICES_DTYPE = np.dtype([("x", "<f8"), ("y", "<f8"), ("z", "<f8")])
    _minimum_vertices = 1

    def __init__(
        self,
        vertices: np.ndarray | list | tuple | None = None,
        **kwargs,
    ):
        self._vertices: np.ndarray = self.validate_vertices(vertices)

        super().__init__(**kwargs)

    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ) -> Points:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.copy`.
        """
        if mask is not None and self.vertices is not None:
            if (
                not isinstance(mask, np.ndarray)
                or mask.shape != (self.n_vertices,)
                or mask.dtype != bool
            ):
                raise ValueError(
                    "Mask must be an array of shape (n_vertices,) and dtype 'bool'."
                )

            kwargs.update({"vertices": self.vertices[mask]})

        new_entity = super().copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            mask=mask,
            **kwargs,
        )

        return new_entity

    @property
    def n_vertices(self) -> int:
        """
        Number of vertices
        """
        return self.vertices.shape[0]

    def remove_vertices(
        self, indices: list[int] | np.ndarray, clear_cache: bool = False
    ):
        """
        Safely remove vertices and corresponding data entries.

        :param indices: Indices of vertices to remove.
        :param clear_cache: Clear cached data and attributes.
        """
        if isinstance(indices, list):
            indices = np.array(indices)

        if not isinstance(indices, np.ndarray):
            raise TypeError("Indices must be a list or numpy array.")

        if np.max(indices) > self.n_vertices - 1:
            raise ValueError("Found indices larger than the number of vertices.")

        if (self.n_vertices - len(np.unique(indices))) < self._minimum_vertices:
            raise ValueError(
                f"Operation would leave fewer vertices than the "
                f"minimum permitted of {self._minimum_vertices}."
            )

        vertices = np.delete(self.vertices, indices, axis=0)
        self.load_children_values()
        self._vertices = self.validate_vertices(vertices)
        self._remove_children_values(
            indices, DataAssociationEnum.VERTEX, clear_cache=clear_cache
        )
        self.workspace.update_attribute(self, "vertices")

    @property
    def vertices(self) -> np.ndarray:
        """
        Array of vertices coordinates, shape(n_vertices, 3).
        """
        if self._vertices is None and self.on_file:
            self._vertices = self.workspace.fetch_array_attribute(self, "vertices")

        return self._vertices.view("<f8").reshape((-1, 3))

    @vertices.setter
    def vertices(self, vertices: np.ndarray | list | tuple):
        xyz = self.validate_vertices(vertices)
        if self._vertices is not None and self._vertices.shape != xyz.shape:
            raise ValueError(
                "New vertices array must have the same shape as the current vertices array."
            )
        self._vertices = xyz

        self.workspace.update_attribute(self, "vertices")

    @classmethod
    def validate_vertices(cls, xyz: np.ndarray | list | tuple | None) -> np.ndarray:
        """
        Validate and format type of vertices array.

        :param xyz: Array of vertices as defined by :obj:`~geoh5py.objects.points.Points.vertices`.
        """
        if xyz is None:
            warnings.warn(
                "No 'vertices' provided. Using (0, 0, 0) default point at the origin.",
                UserWarning,
            )
            xyz = (0.0, 0.0, 0.0)

        if isinstance(xyz, (list, tuple)):
            xyz = np.array(xyz, ndmin=2)

        if not isinstance(xyz, np.ndarray):
            raise TypeError("Vertices must be a numpy array.")

        if len(xyz) < cls._minimum_vertices:
            warnings.warn(
                f"Attribute 'vertices' has fewer elements than the "
                f"minimum required for object of type {type(cls)}. "
                f"Augmenting the array to shape ({cls._minimum_vertices}, 3).",
                UserWarning,
            )
            xyz = np.vstack([xyz] * cls._minimum_vertices)

        if np.issubdtype(xyz.dtype, np.number):
            if xyz.ndim != 2 or xyz.shape[-1] != 3:
                raise ValueError(
                    "Array of 'vertices' should be of shape (*, 3). "
                    f"Got shape {xyz.shape}."
                )

            xyz = np.asarray(
                np.core.records.fromarrays(
                    xyz.T.tolist(),
                    dtype=cls.__VERTICES_DTYPE,
                )
            )

        if xyz.dtype != np.dtype(cls.__VERTICES_DTYPE):
            raise ValueError(
                f"Array of 'vertices' must be of dtype = {cls.__VERTICES_DTYPE}"
            )

        return xyz



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-few-public-methods

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum

from ..data import Data, DataAssociationEnum, FloatData, NumericData


class PropertyGroupType(ABC):
    """
    Class to define the basic structure of a property group type.
    """

    no_modify: bool = False

    @classmethod
    @abstractmethod
    def verify(cls, children: list[Data]):
        """
        Verify that the children are of the correct type for this group type

        If unsuccessful, raise a ValueError

        :param children: The children to verify
        """


class SimpleType(PropertyGroupType):
    @classmethod
    def verify(cls, children: list[Data]):
        """
        Accept any children
        """


class DepthType(PropertyGroupType):
    @classmethod
    def verify(cls, children: list[Data]):
        if (
            len(children) < 1
            or not children[0].association == DataAssociationEnum.DEPTH
            or not isinstance(children[0], FloatData)
        ):
            raise TypeError(
                "First children of 'Depth table' property group type "
                "must be a FloatData of 'Depth' association. "
            )


class DipDirType(PropertyGroupType):
    no_modify = True

    @classmethod
    def verify(cls, children: list[Data]):
        if len(children) != 2 or not all(
            isinstance(child, NumericData) for child in children
        ):
            raise TypeError(
                "Children of 'Dip direction & dip' property group type "
                "must be a list of 2 NumericData entities"
            )


class IntervalType(PropertyGroupType):
    @classmethod
    def verify(cls, children: list[Data]):
        if (
            len(children) < 2
            or not children[0].association == DataAssociationEnum.DEPTH
            or not all(isinstance(child, FloatData) for child in children[:2])
        ):
            raise TypeError(
                "First two children of 'Interval table' property group type "
                "must be FloatData of 'Interval' association."
            )


class MultiElementType(PropertyGroupType):
    @classmethod
    def verify(cls, children: list[Data]):
        if not all(isinstance(child, NumericData) for child in children):
            raise TypeError(
                "Children of 'Multi-element' property group type "
                "must be a list of NumericData entities"
            )


class StrikeDipType(PropertyGroupType):
    no_modify = True

    @classmethod
    def verify(cls, children: list[Data]):
        if len(children) != 2 or not all(
            isinstance(child, NumericData) for child in children
        ):
            raise TypeError(
                "Children of 'Strike & dip' property group type "
                "must be a list of 2 NumericData entities"
            )


class VectorType(PropertyGroupType):
    no_modify = True

    @classmethod
    def verify(cls, children: list[Data]):
        if len(children) != 3 or not all(
            isinstance(child, NumericData) for child in children
        ):
            raise TypeError(
                "Children of '3D vector' property group type "
                "must be a list of 3 NumericData entities"
            )


GROUP_TYPES = {
    "Depth table": DepthType,
    "Dip direction & dip": DipDirType,
    "Interval table": IntervalType,
    "Multi-element": MultiElementType,
    "Simple": SimpleType,
    "Strike & dip": StrikeDipType,
    "3D vector": VectorType,
}


class GroupTypeEnum(str, Enum):
    """
    Supported property group types.
    """

    DEPTH = "Depth table"
    DIPDIR = "Dip direction & dip"
    INTERVAL = "Interval table"
    MULTI = "Multi-element"
    SIMPLE = "Simple"
    STRIKEDIP = "Strike & dip"
    VECTOR = "3D vector"

    @classmethod
    def find_type(cls, data: list[Data]):
        """
        Determine the group type based on the data.
        """
        if all(isinstance(d, NumericData) for d in data):
            return cls.MULTI
        return cls.SIMPLE

    def verify(self, data: list[Data]):
        """
        Validate the data based on the group type.
        """
        GROUP_TYPES[self.value].verify(data)

    @property
    def no_modify(self) -> bool:
        """
        Get the name of the group type.
        """
        return GROUP_TYPES[self.value].no_modify



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import xml.etree.ElementTree as ET

import numpy as np

from .data_association_enum import DataAssociationEnum
from .text_data import TextData


PARAMETERS = ["Colour"]
ATTRIBUTES = ["tag", "text", "attrib"]


class VisualParameters(TextData):
    _xml: ET.Element | None = None
    _association = DataAssociationEnum.OBJECT

    @property
    def xml(self) -> ET.Element:
        """
        :obj:`str` XML string.
        """
        if self._xml is None:
            if isinstance(self.values, str):
                str_xml = self.values
            else:
                str_xml = """
                    <IParameterList Version="1.0">
                    </IParameterList>
                """

            self._xml = ET.fromstring(str_xml)

        return self._xml

    @property
    def values(self) -> str | None:
        if self._xml is not None:
            self._values = ET.tostring(self._xml, encoding="unicode")

        elif (getattr(self, "_values", None) is None) and self.on_file:
            values = self.workspace.fetch_values(self)
            if isinstance(values, (np.ndarray, str, type(None))):
                self._values = values

        return self._values

    @values.setter
    def values(self, values: np.ndarray | str | None):
        if not isinstance(values, (np.ndarray, str, type(None))):
            raise ValueError(
                f"Input 'values' for {self} must be of type {np.ndarray}  str or None."
            )

        self._values = values
        self.workspace.update_attribute(self, "values")

    @property
    def colour(self) -> None | list:
        """
        Colour of the object in [Red, Green, Blue] format.

        Each value is an integer between 0 and 255.
        The colour value is stored as a single integer converted from
        a byte string of the form 'RRGGBB' where 'BB' is the blue value,
        'GG' is the green value, 'RR' is the red converted from hexadecimal format.
        """
        element = self.get_tag("Colour")

        if element is None or not element.text:
            return None

        c_string = (int(element.text)).to_bytes(4, byteorder="little").hex()

        return [int(c_string[i : i + 2], 16) for i in range(0, 8, 2)][:3]

    @colour.setter
    def colour(self, rgb: list | tuple | np.ndarray):
        if (
            not isinstance(rgb, (list, tuple, np.ndarray))
            or len(rgb) != 3
            or not all(isinstance(val, int) for val in rgb)
        ):
            raise TypeError("Input 'colour' values must be a list of 3 integers.")

        byte_string = "".join(f"{val:02x}" for val in rgb)
        byte_string.join(f"{255:02x}")  # alpha value
        value = int.from_bytes(bytes.fromhex(byte_string), "little")

        self.set_tag("Colour", str(value))

    def get_tag(self, tag: str) -> None | ET.Element:
        """
        Recover the tag element.

        :param tag: The name of the tag.

        :return: The xml element.
        """
        element = self.xml.find(tag)
        return element  # type: ignore

    def set_tag(self, tag: str, value: str):
        """
        Set the value for the tag.

        :param tag: The name of the tag.
        :param value: the value to set.
        """

        if not isinstance(value, str):
            raise TypeError(
                f"Input 'value' for VisualParameters.{tag} must be of type {str}."
            )

        if self.xml.find(tag) is None:
            ET.SubElement(self.xml, tag)

        element = self.get_tag(tag)

        if element is not None:
            element.text = value
            self.workspace.update_attribute(self, "values")



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import TYPE_CHECKING

from ...data import Data
from ..utils import as_str_if_uuid
from .concatenated import Concatenated


if TYPE_CHECKING:
    from .object import ConcatenatedObject
    from .property_group import ConcatenatedPropertyGroup


class ConcatenatedData(Concatenated, Data):
    def __init__(self, **kwargs):
        if kwargs.get("parent") is None:
            raise UserWarning(
                "Creating a concatenated data must have a parent "
                "of type Concatenated."
            )

        super().__init__(**kwargs)

    @property
    def property_group(self) -> ConcatenatedPropertyGroup | None:
        """Get the property group containing the data interval."""
        if self.parent.property_groups is None:
            return None

        for prop_group in self.parent.property_groups:
            if prop_group.properties is None:
                continue

            if self.uid in prop_group.properties:
                return prop_group

        return None

    @property
    def parent(self) -> ConcatenatedObject:
        return self._parent

    @parent.setter
    def parent(self, parent: ConcatenatedObject):
        if not hasattr(parent, "add_children"):
            raise ValueError(
                "The 'parent' of a concatenated data must have an 'add_children' method."
            )
        parent.add_children([self])
        self._parent: ConcatenatedObject = parent

        parental_attr = self.concatenator.get_concatenated_attributes(self.parent.uid)

        if f"Property:{self.name}" not in parental_attr:
            parental_attr[f"Property:{self.name}"] = as_str_if_uuid(self.uid)

    @property
    def n_values(self) -> int | None:
        """Number of values in the data."""

        n_values = None
        depths = getattr(self.property_group, "depth_", None)
        if depths and depths is not self:
            n_values = len(depths.values)
        intervals = getattr(self.property_group, "from_", None)
        if intervals and intervals is not self:
            n_values = len(intervals.values)

        return n_values



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import numpy as np

from ...objects import Points
from ...workspace import Workspace
from .base import BaseMerger


class PointsMerger(BaseMerger):
    _type: type = Points

    @classmethod
    def create_object(
        cls, workspace: Workspace, input_entities: list, **kwargs
    ) -> Points:
        # create the vertices
        vertices = np.vstack([input_entity.vertices for input_entity in input_entities])

        # create an object of type
        output = cls._type.create(  # type: ignore
            workspace, vertices=vertices, **kwargs
        )

        return output

    @classmethod
    def validate_structure(cls, input_entity: Points):
        """
        Validate the input entity structure and raises error if incompatible.
        :param input_entity: the input entity to validate.
        """
        # verify if the input entity have vertices
        if not isinstance(input_entity.vertices, np.ndarray):
            raise AttributeError("All entities must have vertices.")
